# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2022 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Module
"""
import datetime
from typing import Any

import pytest
import yaml
from dateutil import tz

import datacube.scripts.search_tool
from datacube.config import LocalConfig
from datacube.drivers.postgres._connections import DEFAULT_DB_USER
from datacube.index import Index
from datacube.model import Dataset
from datacube.model import Product
from datacube.model import Range

from datacube import Datacube
from .search_utils import assume_utc, _cli_csv_search, _csv_search_raw, _load_product_query


def test_search_dataset_equals_eo3(index: Index, ls8_eo3_dataset: Dataset):
    datasets = index.datasets.search_eager(
        platform='landsat-8'
    )
    assert len(datasets) == 1
    assert datasets[0].id == ls8_eo3_dataset.id

    datasets = index.datasets.search_eager(
        platform='landsat-8',
        instrument='OLI_TIRS'
    )
    assert len(datasets) == 1
    assert datasets[0].id == ls8_eo3_dataset.id

    # Wrong product family
    with pytest.raises(ValueError):
        datasets = index.datasets.search_eager(
            platform='landsat-8',
            product_family='splunge',
        )


def test_search_dataset_by_metadata_eo3(index: Index, ls8_eo3_dataset: Dataset) -> None:
    datasets = index.datasets.search_by_metadata(
        {"properties": {"eo:platform": "landsat-8", "eo:instrument": "OLI_TIRS"}}
    )
    datasets = list(datasets)
    assert len(datasets) == 1
    assert datasets[0].id == ls8_eo3_dataset.id

    datasets = index.datasets.search_by_metadata(
        {"properties": {"eo:platform": "landsat-5", "eo:instrument": "TM"}}
    )
    datasets = list(datasets)
    assert len(datasets) == 0


def test_search_day_eo3(index: Index, ls8_eo3_dataset: Dataset) -> None:
    # Matches day
    datasets = index.datasets.search_eager(
        time=datetime.date(2016, 5, 12)
    )
    assert len(datasets) == 1
    assert datasets[0].id == ls8_eo3_dataset.id

    # Different day: no match
    datasets = index.datasets.search_eager(
        time=datetime.date(2016, 5, 13)
    )
    assert len(datasets) == 0


def test_search_dataset_ranges_eo3(index: Index, ls8_eo3_dataset: Dataset) -> None:
    # In the lat bounds.
    datasets = index.datasets.search_eager(
        lat=Range(-37.5, -36.5),
        time=Range(
            datetime.datetime(2016, 5, 12, 23, 0, 0),
            datetime.datetime(2016, 5, 12, 23, 59, 59)
        )
    )
    assert len(datasets) == 1
    assert datasets[0].id == ls8_eo3_dataset.id

    # Out of the lat bounds.
    datasets = index.datasets.search_eager(
        lat=Range(28, 32),
        time=Range(
            datetime.datetime(2016, 5, 12, 23, 0, 0),
            datetime.datetime(2016, 5, 12, 23, 59, 59)
        )
    )
    assert len(datasets) == 0

    # Out of the time bounds
    datasets = index.datasets.search_eager(
        lat=Range(-37.5, -36.5),
        time=Range(
            datetime.datetime(2014, 7, 26, 21, 48, 0),
            datetime.datetime(2014, 7, 26, 21, 50, 0)
        )
    )
    assert len(datasets) == 0

    # A dataset that overlaps but is not fully contained by the search bounds.
    # Should we distinguish between 'contains' and 'overlaps'?
    datasets = index.datasets.search_eager(
        lat=Range(-40, -37.1)
    )
    assert len(datasets) == 1
    assert datasets[0].id == ls8_eo3_dataset.id

    # Single point search
    datasets = index.datasets.search_eager(
        lat=-37.0,
        time=Range(
            datetime.datetime(2016, 5, 12, 23, 0, 0),
            datetime.datetime(2016, 5, 12, 23, 59, 59)
        )
    )
    assert len(datasets) == 1
    assert datasets[0].id == ls8_eo3_dataset.id

    datasets = index.datasets.search_eager(
        lat=30.0,
        time=Range(
            datetime.datetime(2016, 5, 12, 23, 0, 0),
            datetime.datetime(2016, 5, 12, 23, 59, 59)
        )
    )
    assert len(datasets) == 0

    # Single timestamp search
    datasets = index.datasets.search_eager(
        lat=Range(-37.5, -36.5),
        time=datetime.datetime(2016, 5, 12, 23, 50, 40),
    )
    assert len(datasets) == 1
    assert datasets[0].id == ls8_eo3_dataset.id

    datasets = index.datasets.search_eager(
        lat=Range(-37.5, -36.5),
        time=datetime.datetime(2016, 5, 12, 23, 0, 0)
    )
    assert len(datasets) == 0


def test_zero_width_range_search(index: Index, ls8_eo3_dataset4: Dataset) -> None:
    # Test time search against zero-width time metadata
    datasets = index.datasets.search_eager(time=Range(
        begin=datetime.datetime(2013, 7, 21, 0, 57, 26, 432563, tzinfo=datetime.timezone.utc),
        end=datetime.datetime(2013, 7, 21, 0, 57, 26, 432563, tzinfo=datetime.timezone.utc)
    ))
    assert len(datasets) == 1

    datasets = index.datasets.search_eager(time=Range(
        begin=datetime.datetime(2013, 7, 21, 0, 57, 26, 432563, tzinfo=datetime.timezone.utc),
        end=datetime.datetime(2013, 7, 21, 0, 57, 27, 432563, tzinfo=datetime.timezone.utc)
    ))
    assert len(datasets) == 1

    datasets = index.datasets.search_eager(time=Range(
        begin=datetime.datetime(2013, 7, 21, 0, 57, 25, 432563, tzinfo=datetime.timezone.utc),
        end=datetime.datetime(2013, 7, 21, 0, 57, 26, 432563, tzinfo=datetime.timezone.utc)
    ))
    assert len(datasets) == 1


def test_search_globally_eo3(index: Index, ls8_eo3_dataset: Dataset) -> None:
    # No expressions means get all.
    results = list(index.datasets.search())
    assert len(results) == 1

    # Dataset sources aren't loaded by default
    assert results[0].sources is None


def test_search_by_product_eo3(index: Index,
                               base_eo3_product_doc: Product,
                               ls8_eo3_dataset: Dataset,
                               wo_eo3_dataset: Dataset) -> None:
    # Query all the test data, the counts should match expected
    results = _load_product_query(index.datasets.search_by_product())
    assert len(results) == 2
    dataset_count = sum(len(ds) for ds in results.values())
    assert dataset_count == 2

    # Query one product
    products = _load_product_query(index.datasets.search_by_product(
        platform='landsat-8',
        product_family='wo'
    ))
    assert len(products) == 1
    [dataset] = products[base_eo3_product_doc["name"]]
    assert dataset.id == wo_eo3_dataset.id


def test_search_limit_eo3(index, ls8_eo3_dataset, ls8_eo3_dataset2, wo_eo3_dataset):
    prod = ls8_eo3_dataset.type.name
    datasets = list(index.datasets.search(product=prod))
    assert len(datasets) == 2
    datasets = list(index.datasets.search(limit=1, product=prod))
    ids = [ds.id for ds in datasets]
    assert len(ids) == 1
    assert len(datasets) == 1
    datasets = list(index.datasets.search(limit=0, product=prod))
    assert len(datasets) == 0
    datasets = list(index.datasets.search(limit=5, product=prod))
    assert len(datasets) == 2

    datasets = list(index.datasets.search_returning(('id',), product=prod))
    assert len(datasets) == 2
    datasets = list(index.datasets.search_returning(('id',), limit=1, product=prod))
    assert len(datasets) == 1
    datasets = list(index.datasets.search_returning(('id',), limit=0, product=prod))
    assert len(datasets) == 0
    datasets = list(index.datasets.search_returning(('id',), limit=5, product=prod))
    assert len(datasets) == 2

    # Limit is per product not overall.  (But why?!?)
    datasets = list(index.datasets.search())
    assert len(datasets) == 3
    datasets = list(index.datasets.search(limit=1))
    assert len(datasets) == 2
    datasets = list(index.datasets.search(limit=0))
    assert len(datasets) == 0
    datasets = list(index.datasets.search(limit=5))
    assert len(datasets) == 3

    datasets = list(index.datasets.search_returning(('id',)))
    assert len(datasets) == 3
    datasets = list(index.datasets.search_returning(('id',), limit=1))
    assert len(datasets) == 2
    datasets = list(index.datasets.search_returning(('id',), limit=0))
    assert len(datasets) == 0
    datasets = list(index.datasets.search_returning(('id',), limit=5))
    assert len(datasets) == 3


def test_search_or_expressions_eo3(index: Index,
                                   ls8_eo3_dataset: Dataset,
                                   ls8_eo3_dataset2: Dataset,
                                   wo_eo3_dataset: Dataset) -> None:
    # Three EO3 datasets:
    # - two landsat8 ard
    # - one wo

    all_datasets = index.datasets.search_eager()
    assert len(all_datasets) == 3
    all_ids = set(dataset.id for dataset in all_datasets)

    # OR all instruments: should return all datasets
    datasets = index.datasets.search_eager(
        instrument=['WOOLI_TIRS', 'OLI_TIRS', 'OLI_TIRS2']
    )
    assert len(datasets) == 3
    ids = set(dataset.id for dataset in datasets)
    assert ids == all_ids

    # OR expression with only one clause.
    datasets = index.datasets.search_eager(
        instrument=['OLI_TIRS']
    )
    assert len(datasets) == 1
    assert datasets[0].id == ls8_eo3_dataset.id

    # OR both products: return all
    datasets = index.datasets.search_eager(
        product=[ls8_eo3_dataset.type.name, wo_eo3_dataset.type.name]
    )
    assert len(datasets) == 3
    ids = set(dataset.id for dataset in datasets)
    assert ids == all_ids

    # eo OR eo3: return all
    datasets = index.datasets.search_eager(
        metadata_type=[
            # LS5 + children
            ls8_eo3_dataset.metadata_type.name,
            # Nothing
            # LS8 dataset
            wo_eo3_dataset.metadata_type.name
        ]
    )
    assert len(datasets) == 3
    ids = set(dataset.id for dataset in datasets)
    assert ids == all_ids

    # Redundant ORs should have no effect.
    datasets = index.datasets.search_eager(
        product=[wo_eo3_dataset.type.name, wo_eo3_dataset.type.name, wo_eo3_dataset.type.name]
    )
    assert len(datasets) == 1
    assert datasets[0].id == wo_eo3_dataset.id


def test_search_returning_eo3(index: Index,
                              local_config: LocalConfig,
                              ls8_eo3_dataset: Dataset,
                              ls8_eo3_dataset2: Dataset,
                              wo_eo3_dataset: Dataset) -> None:
    assert index.datasets.count() == 3, "Expected three test datasets"

    # Expect one product with our one dataset.
    results = list(index.datasets.search_returning(
        ('id', 'region_code', 'dataset_maturity'),
        platform='landsat-8',
        instrument='OLI_TIRS',
    ))
    assert len(results) == 1
    id_, region_code, maturity = results[0]
    assert id_ == ls8_eo3_dataset.id
    assert region_code == '090086'
    assert maturity == 'final'

    results = list(index.datasets.search_returning(
        ('id', 'metadata_doc',),
        platform='landsat-8',
        instrument='OLI_TIRS',
    ))
    assert len(results) == 1
    id_, document = results[0]
    assert id_ == ls8_eo3_dataset.id
    assert document == ls8_eo3_dataset.metadata_doc

    my_username = local_config.get('db_username', DEFAULT_DB_USER)

    # Mixture of document and native fields
    results = list(index.datasets.search_returning(
        ('id', 'creation_time', 'format', 'label'),
        platform='landsat-8',
        instrument='OLI_TIRS',
        indexed_by=my_username,
    ))
    assert len(results) == 1

    id_, creation_time, format_, label = results[0]

    assert id_ == ls8_eo3_dataset.id
    assert format_ == 'GeoTIFF'

    # It's always UTC in the document
    expected_time = creation_time.astimezone(tz.tzutc()).replace(tzinfo=None)
    assert expected_time.isoformat() == ls8_eo3_dataset.metadata.creation_dt
    assert label == ls8_eo3_dataset.metadata.label


def test_search_returning_rows_eo3(index,
                                   eo3_ls8_dataset_doc,
                                   eo3_ls8_dataset2_doc,
                                   ls8_eo3_dataset, ls8_eo3_dataset2):
    dataset = ls8_eo3_dataset
    uri = eo3_ls8_dataset_doc[1]
    uri3 = eo3_ls8_dataset2_doc[1]
    results = list(index.datasets.search_returning(
        ('id', 'uri'),
        platform='landsat-8',
        instrument='OLI_TIRS',
    ))
    assert len(results) == 1
    assert results == [(dataset.id, uri)]

    index.datasets.archive_location(dataset.id, uri)
    index.datasets.remove_location(dataset.id, uri)

    # If returning a field like uri, there will be one result per location.
    # No locations
    results = list(index.datasets.search_returning(
        ('id', 'uri'),
        platform='landsat-8',
        instrument='OLI_TIRS',
    ))
    assert len(results) == 0

    # Add a second location and we should get two results
    index.datasets.add_location(dataset.id, uri)
    uri2 = 'file:///tmp/test2'
    index.datasets.add_location(dataset.id, uri2)
    results = set(index.datasets.search_returning(
        ('id', 'uri'),
        platform='landsat-8',
        instrument='OLI_TIRS',
    ))
    assert len(results) == 2
    assert results == {
        (dataset.id, uri),
        (dataset.id, uri2)
    }

    # A second dataset already has a location:
    results = set(index.datasets.search_returning(
        ('id', 'uri'),
        platform='landsat-8',
        dataset_maturity='final',
    ))
    assert len(results) == 3
    assert results == {
        (dataset.id, uri),
        (dataset.id, uri2),
        (ls8_eo3_dataset2.id, uri3),
    }


def test_searches_only_type_eo3(index: Index,
                                wo_eo3_dataset: Dataset,
                                ls8_eo3_dataset: Dataset) -> None:
    assert ls8_eo3_dataset.metadata_type.name != wo_eo3_dataset.metadata_type.name

    # One result in the product
    datasets = index.datasets.search_eager(
        product=wo_eo3_dataset.type.name,
        platform='landsat-8'
    )
    assert len(datasets) == 1
    assert datasets[0].id == wo_eo3_dataset.id

    # One result in the metadata type
    datasets = index.datasets.search_eager(
        metadata_type="eo3",
        platform='landsat-8'
    )
    assert len(datasets) == 1
    assert datasets[0].id == wo_eo3_dataset.id

    # No results when searching for a different dataset type.
    with pytest.raises(ValueError):
        datasets = index.datasets.search_eager(
            product="spam_and_eggs",
            platform='landsat-8'
        )

    # Two result when no types specified.
    datasets = index.datasets.search_eager(
        platform='landsat-8'
    )
    assert len(datasets) == 2
    assert set(ds.id for ds in datasets) == {ls8_eo3_dataset.id, wo_eo3_dataset.id}

    # No results for different metadata type.
    with pytest.raises(ValueError):
        datasets = index.datasets.search_eager(
            metadata_type='spam_type',
            platform='landsat-8',
        )


def test_search_special_fields_eo3(index: Index,
                                   ls8_eo3_dataset: Dataset,
                                   wo_eo3_dataset: Dataset) -> None:
    # 'product' is a special case
    datasets = index.datasets.search_eager(
        product=ls8_eo3_dataset.type.name
    )
    assert len(datasets) == 1
    assert datasets[0].id == ls8_eo3_dataset.id

    # Unknown field: no results
    with pytest.raises(ValueError):
        datasets = index.datasets.search_eager(
            platform='landsat-8',
            flavour='vanilla',
        )


def test_search_by_uri_eo3(index, ls8_eo3_dataset, ls8_eo3_dataset2, eo3_ls8_dataset_doc):
    datasets = index.datasets.search_eager(product=ls8_eo3_dataset.type.name,
                                           uri=eo3_ls8_dataset_doc[1])
    assert len(datasets) == 1

    datasets = index.datasets.search_eager(product=ls8_eo3_dataset.type.name,
                                           uri='file:///x/yz')
    assert len(datasets) == 0


def test_search_conflicting_types(index, ls8_eo3_dataset):
    # Should return no results.
    with pytest.raises(ValueError):
        index.datasets.search_eager(
            product=ls8_eo3_dataset.type.name,
            # The ls8 type is not of type storage_unit.
            metadata_type='storage_unit'
        )


def test_fetch_all_of_md_type(index: Index, ls8_eo3_dataset: Dataset) -> None:
    # Get every dataset of the md type.
    assert ls8_eo3_dataset.metadata_type is not None  # to shut up mypy
    results = index.datasets.search_eager(
        metadata_type=ls8_eo3_dataset.metadata_type.name
    )
    assert len(results) == 1
    assert results[0].id == ls8_eo3_dataset.id
    # Get every dataset of the type.
    results = index.datasets.search_eager(
        product=ls8_eo3_dataset.type.name
    )
    assert len(results) == 1
    assert results[0].id == ls8_eo3_dataset.id

    # No results for another.
    with pytest.raises(ValueError):
        results = index.datasets.search_eager(
            metadata_type='spam_and_eggs'
        )


def test_count_searches(index: Index,
                        ls8_eo3_dataset: Dataset) -> None:
    # One result in the telemetry type
    datasets = index.datasets.count(
        product=ls8_eo3_dataset.type.name,
        platform='landsat-8',
        instrument='OLI_TIRS',
    )
    assert datasets == 1

    # One result in the metadata type
    datasets = index.datasets.count(
        metadata_type=ls8_eo3_dataset.metadata_type.name,
        platform='landsat-8',
        instrument='OLI_TIRS',
    )
    assert datasets == 1

    # No results when searching for a different dataset type.
    datasets = index.datasets.count(
        product="spam_and_eggs",
        platform='landsat-8',
        instrument='OLI_TIRS'
    )
    assert datasets == 0

    # One result when no types specified.
    datasets = index.datasets.count(
        platform='landsat-8',
        instrument='OLI_TIRS',
    )
    assert datasets == 1

    # No results for different metadata type.
    datasets = index.datasets.count(
        metadata_type='spam_and_eggs',
        platform='landsat-8',
        instrument='OLI_TIRS'
    )
    assert datasets == 0


def test_count_by_product_searches_eo3(index: Index,
                                       ls8_eo3_dataset: Dataset,
                                       ls8_eo3_dataset2: Dataset,
                                       wo_eo3_dataset: Dataset) -> None:
    # Two result in the ls8 type
    products = tuple(index.datasets.count_by_product(
        product=ls8_eo3_dataset.type.name,
        platform='landsat-8'
    ))
    assert products == ((ls8_eo3_dataset.type, 2),)

    # Two results in the metadata type
    products = tuple(index.datasets.count_by_product(
        metadata_type=ls8_eo3_dataset.metadata_type.name,
        platform='landsat-8',
    ))
    assert products == ((ls8_eo3_dataset.type, 2),)

    # No results when searching for a different dataset type.
    products = tuple(index.datasets.count_by_product(
        product="spam_and_eggs",
        platform='landsat-8'
    ))
    assert products == ()

    # Three results over 2 products when no types specified.
    products = set(index.datasets.count_by_product(
        platform='landsat-8',
    ))
    assert products == {(ls8_eo3_dataset.type, 2), (wo_eo3_dataset.type, 1)}

    # No results for different metadata type.
    products = tuple(index.datasets.count_by_product(
        metadata_type='spam_and_eggs',
    ))
    assert products == ()


def test_count_time_groups(index: Index,
                           ls8_eo3_dataset: Dataset) -> None:
    timeline = list(index.datasets.count_product_through_time(
        '1 day',
        product=ls8_eo3_dataset.type.name,
        time=Range(
            datetime.datetime(2016, 5, 11, tzinfo=tz.tzutc()),
            datetime.datetime(2016, 5, 13, tzinfo=tz.tzutc())
        )
    ))

    assert len(timeline) == 2
    assert timeline == [
        (
            Range(datetime.datetime(2016, 5, 11, tzinfo=tz.tzutc()),
                  datetime.datetime(2016, 5, 12, tzinfo=tz.tzutc())),
            0
        ),
        (
            Range(datetime.datetime(2016, 5, 12, tzinfo=tz.tzutc()),
                  datetime.datetime(2016, 5, 13, tzinfo=tz.tzutc())),
            1
        )
    ]


def test_count_time_groups_cli(clirunner: Any,
                               ls8_eo3_dataset: Dataset) -> None:
    result = clirunner(
        [
            'product-counts',
            '1 day',
            'time in [2016-05-11, 2016-05-13]'
        ], cli_method=datacube.scripts.search_tool.cli,
        verbose_flag=''
    )
    expected_out = (
        f'{ls8_eo3_dataset.type.name}\n'
        '    2016-05-11: 0\n'
        '    2016-05-12: 1\n'
    )
    assert result.output.endswith(expected_out)


def test_search_cli_basic(clirunner: Any,
                          ls8_eo3_dataset: Dataset) -> None:
    """
    Search datasets using the cli.
    """
    result = clirunner(
        [
            # No search arguments: return all datasets.
            'datasets'
        ], cli_method=datacube.scripts.search_tool.cli
    )
    assert str(ls8_eo3_dataset.id) in result.output
    assert str(ls8_eo3_dataset.metadata_type.name) in result.output
    assert result.exit_code == 0


def test_cli_info_eo3(index: Index,
                      clirunner: Any,
                      ls8_eo3_dataset: Dataset,
                      ls8_eo3_dataset2: Dataset,
                      eo3_ls8_dataset_doc) -> None:
    """
    Search datasets using the cli.
    """
    index.datasets.add_location(ls8_eo3_dataset.id, 'file:///tmp/location1')

    opts = [
        'dataset', 'info', str(ls8_eo3_dataset.id)
    ]
    result = clirunner(opts, verbose_flag='')

    output = result.output
    # Remove WARNING messages for experimental driver
    output_lines = [line for line in output.splitlines() if "WARNING:" not in line]
    output = "\n".join(output_lines)

    # Should be a valid yaml
    yaml_docs = list(yaml.safe_load_all(output))
    assert len(yaml_docs) == 1

    # We output properties in order for readability:
    output_lines = set(line for line in output_lines)
    expected_lines = [
        "id: " + str(ls8_eo3_dataset.id),
        'product: ga_ls8c_ard_3',
        'status: active',
        'locations:',
        '- file:///tmp/location1',
        f'- {eo3_ls8_dataset_doc[1]}',
        'fields:',
        '    creation_time: 2019-10-07 20:19:19.218290',
        '    format: GeoTIFF',
        '    instrument: OLI_TIRS',
        '    label: ga_ls8c_ard_3-0-0_090086_2016-05-12_final',
        '    landsat_product_id: LC08_L1TP_090086_20160512_20180203_01_T1',
        '    landsat_scene_id: LC80900862016133LGN02',
        '    lat: {begin: -38.53221689818913, end: -36.41618895501644}',
        '    lon: {begin: 147.65992717003462, end: 150.3003802932316}',
        '    platform: landsat-8',
        '    product_family: ard',
        '    region_code: 090086',
        "    time: {begin: '2016-05-12T23:50:23.054165+00:00', end: '2016-05-12T23:50:52.031499+00:00'}",
    ]
    for line in expected_lines:
        assert line in output_lines

    # Check indexed time separately, as we don't care what timezone it's displayed in.
    indexed_time = yaml_docs[0]['indexed']
    assert isinstance(indexed_time, datetime.datetime)
    assert assume_utc(indexed_time) == assume_utc(ls8_eo3_dataset.indexed_time)

    # Request two, they should have separate yaml documents
    opts.append(str(ls8_eo3_dataset2.id))

    result = clirunner(opts)
    yaml_docs = list(yaml.safe_load_all(result.output))
    assert len(yaml_docs) == 2, "Two datasets should produce two sets of info"
    assert yaml_docs[0]['id'] == str(ls8_eo3_dataset.id)
    assert yaml_docs[1]['id'] == str(ls8_eo3_dataset2.id)


def test_find_duplicates_eo3(index,
                             ls8_eo3_dataset, ls8_eo3_dataset2,
                             ls8_eo3_dataset3, ls8_eo3_dataset4,
                             wo_eo3_dataset):
    # Our four ls8 datasets and one wo.
    all_datasets = index.datasets.search_eager()
    assert len(all_datasets) == 5

    # First two ls8 datasets have the same path/row, last two have a different row.
    expected_ls8_path_row_duplicates = [
        (
            ('090086', 'final'),
            {ls8_eo3_dataset.id, ls8_eo3_dataset2.id}
        ),
        (
            ('101077', 'final'),
            {ls8_eo3_dataset3.id, ls8_eo3_dataset4.id}
        ),

    ]

    # Specifying groups as fields:
    f = ls8_eo3_dataset.metadata_type.dataset_fields.get
    field_res = sorted(index.datasets.search_product_duplicates(
        ls8_eo3_dataset.type,
        f('region_code'), f('dataset_maturity')
    ))
    assert field_res == expected_ls8_path_row_duplicates
    # Field names as strings
    product_res = sorted(index.datasets.search_product_duplicates(
        ls8_eo3_dataset.type,
        'region_code', 'dataset_maturity'
    ))
    assert product_res == expected_ls8_path_row_duplicates

    # No WO duplicates: there's only one
    sat_res = sorted(index.datasets.search_product_duplicates(
        wo_eo3_dataset.type,
        'region_code', 'dataset_maturity'
    ))
    assert sat_res == []


def test_csv_search_via_cli_eo3(clirunner: Any,
                                ls8_eo3_dataset: Dataset,
                                ls8_eo3_dataset2: Dataset) -> None:
    """
    Search datasets via the cli with csv output
    """
    def matches_both(*args):
        rows = _cli_csv_search(('datasets',) + args, clirunner)
        assert len(rows) == 2
        assert {rows[0]['id'], rows[1]['id']} == {str(ls8_eo3_dataset.id), str(ls8_eo3_dataset2.id)}

    def matches_1(*args):
        rows = _cli_csv_search(('datasets',) + args, clirunner)
        assert len(rows) == 1
        assert rows[0]['id'] == str(ls8_eo3_dataset.id)

    def matches_none(*args):
        rows = _cli_csv_search(('datasets',) + args, clirunner)
        assert len(rows) == 0

    def no_such_product(*args):
        with pytest.raises(ValueError):
            _cli_csv_search(('datasets',) + args, clirunner)

    matches_both('lat in [-40, -10]')
    matches_both('product=' + ls8_eo3_dataset.type.name)

    # Don't return on a mismatch
    matches_none('lat in [150, 160]')

    # Match only a single dataset using multiple fields
    matches_1('platform=landsat-8', 'time in [2016-05-11, 2016-05-13]')

    # One matching field, one non-matching
    no_such_product('time in [2016-05-11, 2014-05-13]', 'platform=landsat-5')

    # Test date shorthand
    matches_both('time in [2016-05, 2016-05]')
    matches_none('time in [2014-06, 2014-06]')

    matches_both('time in 2016-05')
    matches_none('time in 2014-08')
    matches_both('time in 2016')
    matches_none('time in 2015')

    matches_both('time in [2016, 2016]')
    matches_both('time in [2015, 2017]')
    matches_none('time in [2015, 2015]')
    matches_none('time in [2013, 2013]')

    matches_both('time in [2016-4, 2016-8]')
    matches_none('time in [2016-1, 2016-3]')
    matches_both('time in [2005, 2017]')


_EXT_AND_BASE_EO3_OUTPUT_HEADER = [
    'id',
    'crs_raw',
    'dataset_maturity',
    'eo_gsd', 'eo_sun_azimuth', 'eo_sun_elevation',
    'cloud_cover', 'fmask_clear', 'fmask_cloud_shadow', 'fmask_snow', 'fmask_water',
    'format',
    'gqa', 'gqa_abs_iterative_mean_x', 'gqa_abs_iterative_mean_xy', 'gqa_abs_iterative_mean_y',
    'gqa_abs_x,gqa_abs_xy', 'gqa_abs_y', 'gqa_cep90',
    'gqa_iterative_mean_x', 'gqa_iterative_mean_xy', 'gqa_iterative_mean_y',
    'gqa_iterative_stddev_x', 'gqa_iterative_stddev_xy', 'gqa_iterative_stddev_y',
    'gqa_mean_x', 'gqa_mean_xy',
    'gqa_mean_y,gqa_stddev_x', 'gqa_stddev_xy', 'gqa_stddev_y',
    'creation_time', 'indexed_by', 'indexed_time',
    'instrument', 'label',
    'landsat_product_id', 'landsat_scene_id',
    'lat', 'lon',
    'metadata_doc', 'metadata_type', 'metadata_type_id',
    'platform', 'product', 'product_family',
    'region_code', 'time', 'uri'
]


def test_csv_structure_eo3(clirunner, ls8_eo3_dataset, ls8_eo3_dataset2):
    output = _csv_search_raw(['datasets', ' lat in [-40, -10]'], clirunner)
    lines = [line.strip() for line in output.split('\n') if line]
    # A header and two dataset rows
    assert len(lines) == 3
    header_line = lines[0]
    for header in _EXT_AND_BASE_EO3_OUTPUT_HEADER:
        assert header in header_line


def test_query_dataset_multi_product_eo3(index: Index, ls8_eo3_dataset, wo_eo3_dataset):
    # We have one ls5 level1 and its child nbar
    dc = Datacube(index)

    # Can we query a single product name?
    datasets = dc.find_datasets(product="ga_ls8c_ard_3")
    assert len(datasets) == 1

    # Can we query multiple products?
    datasets = dc.find_datasets(product=['ga_ls8c_ard_3', 'ga_ls_wo_3'])
    assert len(datasets) == 2

    # Can we query multiple products in a tuple
    datasets = dc.find_datasets(product=('ga_ls8c_ard_3', 'ga_ls_wo_3'))
    assert len(datasets) == 2



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import pytest
from datacube.utils.changes import (
    contains,
    classify_changes,
    allow_any,
    allow_removal,
    allow_addition,
    allow_extension,
    allow_truncation,
    MISSING,
)


def test_changes_contains():
    assert contains("bob", "BOB") is True
    assert contains("bob", "BOB", case_sensitive=True) is False
    assert contains(1, 1) is True
    assert contains(1, {}) is False
    # same as above, but with None interpreted as {}
    assert contains(1, None) is False
    assert contains({}, 1) is False
    assert contains(None, 1) is False
    assert contains({}, {}) is True
    assert contains({}, None) is True

    # this one is arguable...
    assert contains(None, {}) is False
    assert contains(None, None) is True
    assert contains({"a": 1, "b": 2}, {"a": 1}) is True
    assert contains({"a": {"b": "BOB"}}, {"a": {"b": "bob"}}) is True
    assert (
        contains({"a": {"b": "BOB"}}, {"a": {"b": "bob"}}, case_sensitive=True) is False
    )
    assert contains("bob", "alice") is False
    assert contains({"a": 1}, {"a": 1, "b": 2}) is False
    assert contains({"a": {"b": 1}}, {"a": {}}) is True
    assert contains({"a": {"b": 1}}, {"a": None}) is True


def test_classify_changes():
    assert classify_changes([], {}) == ([], [])
    assert classify_changes([(('a',), 1, 2)], {}) == ([], [(('a',), 1, 2)])
    assert classify_changes([(('a',), 1, 2)], {('a',): allow_any}) == ([(('a',), 1, 2)], [])

    changes = [(('a2',), {'b1': 1}, MISSING)]  # {'a1': 1, 'a2': {'b1': 1}} → {'a1': 1}
    good_change = (changes, [])
    bad_change = ([], changes)
    assert classify_changes(changes, {}) == bad_change
    assert classify_changes(changes, {tuple(): allow_any}) == good_change
    assert classify_changes(changes, {tuple(): allow_removal}) == bad_change
    assert classify_changes(changes, {tuple(): allow_addition}) == bad_change
    assert classify_changes(changes, {tuple(): allow_truncation}) == good_change
    assert classify_changes(changes, {tuple(): allow_extension}) == bad_change
    assert classify_changes(changes, {('a1', ): allow_any}) == bad_change
    assert classify_changes(changes, {('a1', ): allow_removal}) == bad_change
    assert classify_changes(changes, {('a1', ): allow_addition}) == bad_change
    assert classify_changes(changes, {('a1', ): allow_truncation}) == bad_change
    assert classify_changes(changes, {('a1', ): allow_extension}) == bad_change
    assert classify_changes(changes, {('a2', ): allow_any}) == good_change
    assert classify_changes(changes, {('a2', ): allow_removal}) == good_change
    assert classify_changes(changes, {('a2', ): allow_addition}) == bad_change
    assert classify_changes(changes, {('a2', ): allow_truncation}) == bad_change
    assert classify_changes(changes, {('a2', ): allow_extension}) == bad_change

    with pytest.raises(RuntimeError):
        classify_changes(changes, {('a2', ): object()})

    assert str(MISSING) == repr(MISSING)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import pytest
from sqlalchemy.exc import OperationalError
from sqlalchemy.engine.url import URL

from datacube.drivers.postgres._connections import PostgresDb, handle_dynamic_token_authentication


counter = [0]
last_base = [None]


def next_token(base):
    counter[0] = counter[0] + 1
    last_base[0] = base
    return f"{base}{counter[0]}"


def test_dynamic_password():
    url = URL.create('postgresql',
                     host="fake_host", database="fake_database", port=6543,
                     username="fake_username", password="fake_password")
    engine = PostgresDb._create_engine(url)
    counter[0] = 0
    last_base[0] = None
    handle_dynamic_token_authentication(engine, next_token, base="password")
    with pytest.raises(OperationalError):
        conn = engine.connect()
    assert counter[0] == 1
    assert last_base[0] == "password"



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import logging

import click
from click import echo, style
from sqlalchemy.exc import OperationalError

import datacube
from datacube.index import index_connect
from datacube.drivers.postgres._connections import IndexSetupError
from datacube.ui import click as ui
from datacube.ui.click import cli, handle_exception
from datacube.config import LocalConfig

_LOG = logging.getLogger('datacube-system')


@cli.group(name='system', help='System commands')
def system():
    pass


@system.command('init', help='Initialise the database')
@click.option(
    '--default-types/--no-default-types', is_flag=True, default=True,
    help="Add default types? (default: true)"
)
@click.option(
    '--init-users/--no-init-users', is_flag=True, default=True,
    help="Include user roles and grants. (default: true)"
)
@click.option(
    '--recreate-views/--no-recreate-views', is_flag=True, default=True,
    help="Recreate dynamic views"
)
@click.option(
    '--rebuild/--no-rebuild', is_flag=True, default=False,
    help="Rebuild all dynamic fields (caution: slow)"
)
@click.option(
    '--lock-table/--no-lock-table', is_flag=True, default=False,
    help="Allow table to be locked (eg. while creating missing indexes)"
)
@ui.pass_index(expect_initialised=False)
def database_init(index, default_types, init_users, recreate_views, rebuild, lock_table):
    echo('Initialising database...')

    was_created = index.init_db(with_default_types=default_types,
                                with_permissions=init_users)

    if was_created:
        echo(style('Created.', bold=True))
    else:
        echo(style('Updated.', bold=True))

    echo('Checking indexes/views.')
    index.metadata_types.check_field_indexes(
        allow_table_lock=lock_table,
        rebuild_indexes=rebuild,
        rebuild_views=recreate_views or rebuild,
    )
    echo('Done.')


@system.command('check', help='Check and display current configuration')
@ui.pass_config
def check(local_config: LocalConfig):
    """
    Verify & view current configuration
    """

    def echo_field(name, value):
        echo('{:<15}'.format(name + ':') + style(str(value), bold=True))

    echo_field('Version', datacube.__version__)
    echo_field('Config files', ','.join(local_config.files_loaded))
    echo_field('Host',
               '{}:{}'.format(local_config['db_hostname'] or 'localhost', local_config.get('db_port', None) or '5432'))

    echo_field('Database', local_config['db_database'])
    echo_field('User', local_config['db_username'])
    echo_field('Environment', local_config['env'])
    echo_field('Index Driver', local_config['index_driver'])

    echo()
    echo('Valid connection:\t', nl=False)
    try:
        index = index_connect(local_config=local_config)
        echo(style('YES', bold=True))
        for role, user, description in index.users.list_users():
            if user == local_config['db_username']:
                echo('You have %s privileges.' % style(role.upper(), bold=True))
    except OperationalError as e:
        handle_exception('Error Connecting to Database: %s', e)
    except IndexSetupError as e:
        handle_exception('Database not initialised: %s', e)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
""" xarray 3D driver plugin for 3D support testing
"""
from contextlib import contextmanager
from pathlib import Path
from typing import Any, Generator, Optional, Tuple, Union
from urllib.parse import urlparse

import numpy as np
import xarray as xr
from affine import Affine
from datacube.storage import BandInfo
from datacube.utils import geometry
from datacube.utils.math import num2numpy

PROTOCOL = ["file", "xarray_3d"]
FORMAT = "xarray_3d"


def uri_split(uri: str) -> Tuple[str, str, str]:
    """
    Splits uri into protocol, root, and group

    Example:
        uri_split('file:///path/to/my_dataset.xarray_3d#group/subgroup/etc')
        returns ('file', '/path/to/my_dataset.xarray_3d', 'group/subgroup/etc')

    If the URI contains no '#' extension, the root group "" is returned.

    :param str uri: The URI to be parsed
    :return: (protocol, root, group)
    """
    components = urlparse(uri)
    scheme = components.scheme
    path = components.netloc + components.path
    if not scheme:
        raise ValueError(f"uri scheme not found: {uri}")
    group = components.fragment
    return scheme, path, group


RasterShape = Tuple[int, ...]
RasterWindow = Tuple[Union[int, Tuple[int, int]], ...]

load_no = 0


class XArrayDataSource3D(object):
    class BandDataSource(object):
        def __init__(
            self,
            dataset: xr.Dataset,
            var_name: str,
            no_data: Optional[float],
        ) -> None:
            """
            Initialises the BandDataSource class.

            The BandDataSource class to read array slices out of the xr.Dataset.

            :param xr.Dataset dataset: The xr.Dataset
            :param str var_name: The variable name of the xr.DataArray
            :param float no_data: The no data value if known
            """
            self.ds = dataset
            self._var_name = var_name
            self.da = dataset.data_vars[var_name]

            self._is_2d = len(self.da.dims) == 2
            self._nbands = 1 if self._is_2d else self.da[self.da.dims[0]].size
            if self._nbands == 0:
                raise ValueError("Dataset has 0 bands.")

            # Set nodata value
            if "nodata" in self.da.attrs and self.da.nodata:
                if isinstance(self.da.nodata, list):
                    self._nodata = self.da.nodata[0]
                else:
                    self._nodata = self.da.nodata
            else:
                self._nodata = no_data

            if not self._nodata:
                raise ValueError("nodata not found in dataset and product definition")

            self._nodata = num2numpy(self._nodata, self.dtype)

        @property
        def nodata(self) -> Optional[float]:
            return self._nodata  # type: ignore

        @property
        def crs(self) -> geometry.CRS:
            return self.da.crs

        @property
        def transform(self) -> Affine:
            return self.da.affine

        @property
        def dtype(self) -> np.dtype:
            return self.da.dtype

        @property
        def shape(self) -> RasterShape:
            return self.da.shape if self._is_2d else self.da.shape[1:]

        def read(
            self,
            window: Optional[RasterWindow] = None,
            out_shape: Optional[RasterShape] = None,
        ) -> np.ndarray:
            """
            Reads a slice into the xr.DataArray.

            :param RasterWindow window: The slice to read
            :param RasterShape out_shape: The desired output shape
            :return: Requested data in a :class:`numpy.ndarray`
            """

            if window is None:
                ix: Tuple = (...,)
            else:
                ix = tuple(slice(*w) if isinstance(w, tuple) else w for w in window)

            def fn() -> Any:
                return self.da[ix].values

            data = fn()

            if out_shape and data.shape != out_shape:
                raise ValueError(
                    f"Data shape does not match 'out_shape': {data.shape} != {out_shape}"
                )

            return data

    def __init__(self, band: BandInfo) -> None:
        """
        Initialises the XarrayDataSource3D class.

        :param BandInfo band: BandInfo containing the dataset metadata.
        """
        self._band_info = band
        if band.band == 0:
            raise ValueError("BandInfo.band must be > 0")

        # convert band.uri -> protocol, root and group
        protocol, filepath, _ = uri_split(band.uri)
        if protocol not in PROTOCOL:
            valid = " or ".join([f"{protocol}://" for protocol in PROTOCOL])
            raise ValueError(f"Expected {valid} url")
        self.filepath = Path(filepath)
        if not self.filepath.exists():
            raise ValueError(f"xarray_3d data file does not exist: {self.filepath}")

    @contextmanager
    def open(self) -> Generator[BandDataSource, None, None]:
        """Load xarray_3d dataset."""
        global load_no

        dataset = xr.open_dataset(self.filepath)

        var_name = self._band_info.layer or self._band_info.name
        load_no += 1
        yield XArrayDataSource3D.BandDataSource(
            dataset=dataset,
            var_name=var_name,
            no_data=self._band_info.nodata,
        )


class XArrayReaderDriver3D(object):
    def __init__(self) -> None:
        self.name = "XArrayReader3D"
        self.protocols = PROTOCOL
        self.formats = [FORMAT]

    def supports(self, protocol: str, fmt: str) -> bool:
        return protocol in self.protocols and fmt in self.formats

    def new_datasource(self, band: BandInfo) -> XArrayDataSource3D:
        return XArrayDataSource3D(band)


def reader_driver_init() -> XArrayReaderDriver3D:
    return XArrayReaderDriver3D()



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Test utility functions from :module:`datacube.utils`


"""
import os
import pathlib
import string
from pathlib import Path

import numpy as np
import pytest
import xarray as xr
from dateutil.parser import parse
from hypothesis import given
from hypothesis.strategies import integers, text
from pandas import to_datetime

from datacube.utils import gen_password, write_user_secret_file, slurp
from datacube.model.utils import xr_apply
from datacube.utils.dates import date_sequence
from datacube.utils.math import (
    num2numpy,
    is_almost_int,
    maybe_zero,
    maybe_int,
    snap_scale,
    valid_mask,
    invalid_mask,
    clamp,
    unsqueeze_data_array,
    unsqueeze_dataset,
    spatial_dims,
    data_resolution_and_offset,
    affine_from_axis,
)
from datacube.utils.py import sorted_items
from datacube.utils.uris import (uri_to_local_path, mk_part_uri, get_part_from_uri, as_url, is_url,
                                 pick_uri, uri_resolve, is_vsipath,
                                 normalise_path, default_base_dir)
from datacube.utils.io import check_write_path
from datacube.testutils import mk_sample_product


def test_stats_dates():
    # Winter for 1990
    winter_1990 = list(date_sequence(start=to_datetime('1990-06-01'), end=to_datetime('1990-09-01'), step_size='3m',
                                     stats_duration='3m'))
    assert winter_1990 == [(parse('1990-06-01'), parse('1990-09-01'))]

    # Every winter from 1990 - 1992
    three_years_of_winter = list(date_sequence(start=to_datetime('1990-06-01'), end=to_datetime('1992-09-01'),
                                               step_size='1y',
                                               stats_duration='3m'))
    assert three_years_of_winter == [(parse('1990-06-01'), parse('1990-09-01')),
                                     (parse('1991-06-01'), parse('1991-09-01')),
                                     (parse('1992-06-01'), parse('1992-09-01'))]

    # Full years from 1990 - 1994
    five_full_years = list(date_sequence(start=to_datetime('1990-01-01'), end=to_datetime('1995'), step_size='1y',
                                         stats_duration='1y'))
    assert five_full_years == [(parse('1990-01-01'), parse('1991-01-01')),
                               (parse('1991-01-01'), parse('1992-01-01')),
                               (parse('1992-01-01'), parse('1993-01-01')),
                               (parse('1993-01-01'), parse('1994-01-01')),
                               (parse('1994-01-01'), parse('1995-01-01'))]

    # Every season (three months), starting in March, from 1990 until end 1992-02
    two_years_of_seasons = list(date_sequence(start=to_datetime('1990-03-01'), end=to_datetime('1992-03'),
                                              step_size='3m',
                                              stats_duration='3m'))
    assert len(two_years_of_seasons) == 8
    assert two_years_of_seasons == [(parse('1990-03-01'), parse('1990-06-01')),
                                    (parse('1990-06-01'), parse('1990-09-01')),
                                    (parse('1990-09-01'), parse('1990-12-01')),
                                    (parse('1990-12-01'), parse('1991-03-01')),
                                    (parse('1991-03-01'), parse('1991-06-01')),
                                    (parse('1991-06-01'), parse('1991-09-01')),
                                    (parse('1991-09-01'), parse('1991-12-01')),
                                    (parse('1991-12-01'), parse('1992-03-01'))]  # Leap year!

    # Every month from 1990-01 to 1990-06
    monthly = list(date_sequence(start=to_datetime('1990-01-01'), end=to_datetime('1990-07-01'), step_size='1m',
                                 stats_duration='1m'))
    assert len(monthly) == 6

    # Complex
    # I want the average over 5 years


def test_uri_to_local_path():
    if os.name == 'nt':
        assert 'C:\\tmp\\test.tmp' == str(uri_to_local_path('file:///C:/tmp/test.tmp'))
        assert '\\\\remote\\path\\file.txt' == str(uri_to_local_path('file://remote/path/file.txt'))

    else:
        assert '/tmp/something.txt' == str(uri_to_local_path('file:///tmp/something.txt'))

        with pytest.raises(ValueError):
            uri_to_local_path('file://remote/path/file.txt')

    assert uri_to_local_path(None) is None

    with pytest.raises(ValueError):
        uri_to_local_path('ftp://example.com/tmp/something.txt')


@pytest.mark.parametrize("base", [
    "s3://foo",
    "gs://foo",
    "wasb://foo",
    "wasbs://foo",
    "/vsizip//vsicurl/https://host.tld/some/path",
])
def test_uri_resolve(base):
    abs_path = '/abs/path/to/something'
    some_uri = 'http://example.com/file.txt'

    assert uri_resolve(base, abs_path) == "file://" + abs_path
    assert uri_resolve(base, some_uri) is some_uri
    assert uri_resolve(base, None) is base
    assert uri_resolve(base, '') is base
    assert uri_resolve(base, 'relative/path') == base + '/relative/path'
    assert uri_resolve(base + '/', 'relative/path') == base + '/relative/path'
    assert uri_resolve(base + '/some/dir/', 'relative/path') == base + '/some/dir/relative/path'

    if not is_vsipath(base):
        assert uri_resolve(base + '/some/dir/file.txt', 'relative/path') == base + '/some/dir/relative/path'


def test_pick_uri():
    f, s, h = ('file://a', 's3://b', 'http://c')

    assert pick_uri([f, s, h]) is f
    assert pick_uri([s, h, f]) is f
    assert pick_uri([s, h]) is s
    assert pick_uri([h, s]) is h
    assert pick_uri([f, s, h], 'http:') is h
    assert pick_uri([f, s, h], 's3:') is s
    assert pick_uri([f, s, h], 'file:') is f

    with pytest.raises(ValueError):
        pick_uri([])

    with pytest.raises(ValueError):
        pick_uri([f, s, h], 'ftp:')

    with pytest.raises(ValueError):
        pick_uri([s, h], 'file:')


@given(integers(), integers(), integers())
def test_clamp(x, lower_bound, upper_bound):
    if lower_bound > upper_bound:
        lower_bound, upper_bound = upper_bound, lower_bound
    new_x = clamp(x, lower_bound, upper_bound)

    # If x was already between the bounds, it shouldn't have changed
    if lower_bound <= x <= upper_bound:
        assert new_x == x
    assert lower_bound <= new_x <= upper_bound


@given(integers(min_value=10, max_value=30))
def test_gen_pass(n_bytes):
    password1 = gen_password(n_bytes)
    password2 = gen_password(n_bytes)
    assert len(password1) >= n_bytes
    assert len(password2) >= n_bytes
    assert password1 != password2


@given(text(alphabet=string.digits + string.ascii_letters + ' ,:.![]?', max_size=20))
def test_write_user_secret_file(txt):
    fname = u".tst-datacube-uefvwr4cfkkl0ijk.txt"

    write_user_secret_file(txt, fname)
    txt_back = slurp(fname)
    os.remove(fname)
    assert txt == txt_back
    assert slurp(fname) is None


def test_testutils_mk_sample():
    pp = mk_sample_product('tt', measurements=[('aa', 'int16', -999),
                                               ('bb', 'float32', np.nan)])
    assert set(pp.measurements) == {'aa', 'bb'}

    pp = mk_sample_product('tt', measurements=['aa', 'bb'])
    assert set(pp.measurements) == {'aa', 'bb'}

    pp = mk_sample_product('tt', measurements=[dict(name=n) for n in ['aa', 'bb']])
    assert set(pp.measurements) == {'aa', 'bb'}

    with pytest.raises(ValueError):
        mk_sample_product('tt', measurements=[None])


def test_testutils_write_files():
    from datacube.testutils import write_files, assert_file_structure

    files = {'a.txt': 'string',
             'aa.txt': ('line1\n', 'line2\n')}

    pp = write_files(files)
    assert pp.exists()
    assert_file_structure(pp, files)

    # test that we detect missing files
    (pp / 'a.txt').unlink()

    with pytest.raises(AssertionError):
        assert_file_structure(pp, files)

    with pytest.raises(AssertionError):
        assert_file_structure(pp, {'aa.txt': 3})

    with pytest.raises(ValueError):
        write_files({'tt': 3})


def test_part_uri():
    base = 'file:///foo.txt'

    for i in range(10):
        assert get_part_from_uri(mk_part_uri(base, i)) == i

    assert get_part_from_uri('file:///f.txt') is None
    assert get_part_from_uri('file:///f.txt#something_else') is None
    assert get_part_from_uri('file:///f.txt#part=aa') == 'aa'
    assert get_part_from_uri('file:///f.txt#part=111') == 111


def test_xr_apply():
    src = xr.DataArray(np.asarray([1, 2, 3], dtype='uint8'), dims=['time'])
    dst = xr_apply(src, lambda _, v: v, dtype='float32')

    assert dst.dtype.name == 'float32'
    assert dst.shape == src.shape
    assert dst.values.tolist() == [1, 2, 3]

    dst = xr_apply(src, lambda _, v: v)
    assert dst.dtype.name == 'uint8'
    assert dst.shape == src.shape
    assert dst.values.tolist() == [1, 2, 3]

    dst = xr_apply(src, lambda idx, _, v: idx[0] + v, with_numeric_index=True)
    assert dst.dtype.name == 'uint8'
    assert dst.shape == src.shape
    assert dst.values.tolist() == [0 + 1, 1 + 2, 2 + 3]


def test_sorted_items():
    aa = dict(c=1, b={}, a=[])

    assert ''.join(k for k, _ in sorted_items(aa)) == 'abc'
    assert ''.join(k for k, _ in sorted_items(aa, key=lambda x: x)) == 'abc'
    assert ''.join(k for k, _ in sorted_items(aa, reverse=True)) == 'cba'

    remap = dict(c=0, a=1, b=2)
    assert ''.join(k for k, _ in sorted_items(aa, key=lambda x: remap[x])) == 'cab'

    assert sorted_items(None) == []


def test_default_base_dir(monkeypatch):
    def set_pwd(p):
        if p is None:
            monkeypatch.delenv('PWD')
        else:
            monkeypatch.setenv('PWD', str(p))

    cwd = Path('.').resolve()

    # Default base dir (once resolved) will never be different from cwd
    assert default_base_dir().resolve() == cwd

    # should work when PWD is not set
    set_pwd(None)
    assert 'PWD' not in os.environ
    assert default_base_dir() == cwd

    # should work when PWD is not absolute path
    set_pwd('this/is/not/a/valid/path')
    assert default_base_dir() == cwd

    # should be cwd when PWD points to some other dir
    set_pwd(cwd / 'deeper')
    assert default_base_dir() == cwd

    set_pwd(cwd.parent)
    assert default_base_dir() == cwd

    # PWD == cwd
    set_pwd(cwd)
    assert default_base_dir() == cwd

    # TODO:
    # - create symlink to current directory in temp
    # - set PWD to that link
    # - make sure that returned path is the same as symlink and different from cwd


def test_time_info():
    from datacube.model.utils import time_info
    from datetime import datetime

    date = '2019-03-03T00:00:00'
    ee = time_info(datetime(2019, 3, 3))
    assert ee['extent']['from_dt'] == date
    assert ee['extent']['to_dt'] == date
    assert ee['extent']['center_dt'] == date
    assert len(ee['extent']) == 3

    ee = time_info(datetime(2019, 3, 3), key_time=datetime(2019, 4, 4))
    assert ee['extent']['from_dt'] == date
    assert ee['extent']['to_dt'] == date
    assert ee['extent']['center_dt'] == date
    assert ee['extent']['key_time'] == '2019-04-04T00:00:00'
    assert len(ee['extent']) == 4


def test_normalise_path():
    cwd = Path('.').resolve()
    assert normalise_path('.').resolve() == cwd

    p = Path('/a/b/c/d.txt')
    assert normalise_path(p) == Path(p)
    assert normalise_path(str(p)) == Path(p)

    base = Path('/a/b/')
    p = Path('c/d.txt')
    assert normalise_path(p, base) == (base / p)
    assert normalise_path(str(p), str(base)) == (base / p)
    assert normalise_path(p) == (cwd / p)

    with pytest.raises(ValueError):
        normalise_path(p, 'not/absolute/path')


def test_testutils_testimage():
    from datacube.testutils import mk_test_image, split_test_image

    for dtype in ('uint16', 'uint32', 'int32', 'float32'):
        aa = mk_test_image(128, 64, dtype=dtype, nodata=None)
        assert aa.shape == (64, 128)
        assert aa.dtype == dtype

        xx, yy = split_test_image(aa)
        assert (xx[:, 33] == 33).all()
        assert (xx[:, 127] == 127).all()
        assert (yy[23, :] == 23).all()
        assert (yy[63, :] == 63).all()


def test_testutils_gtif(tmpdir):
    from datacube.testutils import mk_test_image
    from datacube.testutils.io import write_gtiff, rio_slurp

    w, h, dtype, nodata, ndw = 96, 64, 'int16', -999, 7

    aa = mk_test_image(w, h, dtype, nodata, nodata_width=ndw)
    bb = mk_test_image(w, h, dtype, nodata=None)

    assert aa.shape == (h, w)
    assert aa.dtype.name == dtype
    assert aa[10, 30] == (30 << 8) | 10
    assert aa[10, 11] == nodata
    assert bb[10, 11] == (11 << 8) | 10

    aa5 = np.stack((aa,) * 5)

    fname = pathlib.Path(str(tmpdir / "aa.tiff"))
    fname5 = pathlib.Path(str(tmpdir / "aa5.tiff"))

    aa_meta = write_gtiff(fname, aa, nodata=nodata,
                          blocksize=128,
                          resolution=(100, -100),
                          offset=(12300, 11100),
                          overwrite=True)

    aa5_meta = write_gtiff(str(fname5), aa5, nodata=nodata,
                           resolution=(100, -100),
                           offset=(12300, 11100),
                           overwrite=True)

    assert fname.exists()
    assert fname5.exists()

    assert aa_meta.gbox.shape == (h, w)
    assert aa_meta.path is fname

    aa_, aa_meta_ = rio_slurp(fname)
    aa5_, aa5_meta_ = rio_slurp(fname5)

    assert aa_meta_.path is fname

    (sx, _, tx,
     _, sy, ty, *_) = aa5_meta_.transform

    assert (tx, ty) == (12300, 11100)
    assert (sx, sy) == (100, -100)

    np.testing.assert_array_equal(aa, aa_)
    np.testing.assert_array_equal(aa5, aa5_)

    assert aa_meta_.transform == aa_meta.transform
    assert aa5_meta_.transform == aa5_meta.transform

    # check that overwrite is off by default
    with pytest.raises(IOError):
        write_gtiff(fname, aa, nodata=nodata,
                    blocksize=128)

    # check that overwrite re-writes file
    write_gtiff(fname, bb[:32, :32],
                gbox=aa_meta.gbox[:32, :32],
                overwrite=True)

    bb_, mm = rio_slurp(fname, (32, 32))
    np.testing.assert_array_equal(bb[:32, :32], bb_)

    assert mm.gbox == aa_meta.gbox[:32, :32]

    with pytest.raises(ValueError):
        write_gtiff(fname, np.zeros((3, 4, 5, 6)))


def test_testutils_geobox():
    from datacube.testutils.io import dc_crs_from_rio, rio_geobox
    from rasterio.crs import CRS
    from affine import Affine

    assert rio_geobox({}) is None

    transform = Affine(10, 0, 4676,
                       0, -10, 171878)

    shape = (100, 640)
    h, w = shape
    crs = CRS.from_epsg(3578)

    meta = dict(width=w, height=h, transform=transform, crs=crs)
    gbox = rio_geobox(meta)

    assert gbox.shape == shape
    assert gbox.crs.epsg == 3578
    assert gbox.transform == transform

    wkt = '''PROJCS["unnamed",
    GEOGCS["NAD83",
       DATUM["North_American_Datum_1983",
             SPHEROID["GRS 1980",6378137,298.257222101, AUTHORITY["EPSG","7019"]],
             TOWGS84[0,0,0,0,0,0,0],AUTHORITY["EPSG","6269"]],
       PRIMEM["Greenwich",0,AUTHORITY["EPSG","8901"]],
       UNIT["degree",0.0174532925199433,AUTHORITY["EPSG","9122"]],
       ],
    PROJECTION["Albers_Conic_Equal_Area"],
    PARAMETER["standard_parallel_1",61.66666666666666],
    PARAMETER["standard_parallel_2",68],
    PARAMETER["latitude_of_center",59],
    PARAMETER["longitude_of_center",-132.5],
    PARAMETER["false_easting",500000],
    PARAMETER["false_northing",500000],
    UNIT["Meter",1]]
    '''

    crs_ = dc_crs_from_rio(CRS.from_wkt(wkt))
    assert crs_.epsg is None


@pytest.mark.parametrize("test_input,expected", [
    ("/foo/bar/file.txt", False),
    ("file:///foo/bar/file.txt", True),
    ("test.bar", False),
    ("s3://mybucket/objname.tiff", True),
    ("gs://mybucket/objname.tiff", True),
    ("wasb://mybucket/objname.tiff", True),
    ("wasbs://mybucket/objname.tiff", True),
    ("ftp://host.name/filename.txt", True),
    ("https://host.name.com/path/file.txt", True),
    ("http://host.name.com/path/file.txt", True),
    ("sftp://user:pass@host.name.com/path/file.txt", True),
    ("file+gzip://host.name.com/path/file.txt", True),
    ("bongo:host.name.com/path/file.txt", False),
])
def test_is_url(test_input, expected):
    assert is_url(test_input) == expected
    if expected:
        assert as_url(test_input) is test_input


def test_is_almost_int():
    assert is_almost_int(1, 1e-10)
    assert is_almost_int(1.001, .1)
    assert is_almost_int(2 - 0.001, .1)
    assert is_almost_int(-1.001, .1)


def test_maybe_zero():
    assert maybe_zero(0.0001, 0.1) == 0
    assert maybe_zero(-0.0001, 0.1) == 0
    assert maybe_zero(1.5, 0.1) == 1.5


def test_maybe_int():
    assert maybe_int(1, 1e-10) == 1
    assert maybe_int(1.6, .1) == 1.6
    assert maybe_int(-1.6, .1) == -1.6
    assert maybe_int(1.001, .1) == 1
    assert maybe_int(2 - 0.001, .1) == 2
    assert maybe_int(-1.001, .1) == -1
    assert maybe_int(1.1, .1) == 1.1
    for x in [3/7, 7/3, -13.7878]:
        assert maybe_int(x, 1e-10) is x


def test_snap_scale():
    assert snap_scale(0.9999999) == 1
    assert snap_scale(-0.9999999) == -1
    for x in [0.0, 0.999, 0.621612621868, 3/7, 7/3]:
        assert snap_scale(x) is x
        x = -x
        assert snap_scale(x) is x
    assert snap_scale(0.33333331) == 1/3
    assert snap_scale(-0.33333331) == -1/3


def test_valid_mask():
    xx = np.zeros((4, 8), dtype='float32')
    mm = valid_mask(xx, 0)
    assert mm.dtype == 'bool'
    assert mm.shape == xx.shape
    assert not mm.all()
    assert not mm.any()
    nn = invalid_mask(xx, 0)
    assert nn.dtype == 'bool'
    assert nn.shape == xx.shape
    assert nn.all()
    assert nn.any()

    mm = valid_mask(xx, 13)
    assert mm.dtype == 'bool'
    assert mm.shape == xx.shape
    assert mm.all()
    nn = invalid_mask(xx, 13)
    assert nn.dtype == 'bool'
    assert nn.shape == xx.shape
    assert not nn.any()

    mm = valid_mask(xx, None)
    assert mm.dtype == 'bool'
    assert mm.shape == xx.shape
    assert mm.all()
    nn = invalid_mask(xx, None)
    assert nn.dtype == 'bool'
    assert nn.shape == xx.shape
    assert not nn.any()

    mm = valid_mask(xx, np.nan)
    assert mm.dtype == 'bool'
    assert mm.shape == xx.shape
    assert mm.all()
    nn = invalid_mask(xx, np.nan)
    assert nn.dtype == 'bool'
    assert nn.shape == xx.shape
    assert not nn.any()

    xx[0, 0] = np.nan
    mm = valid_mask(xx, np.nan)
    assert not mm[0, 0]
    assert mm.sum() == (4 * 8 - 1)
    nn = invalid_mask(xx, np.nan)
    assert nn[0, 0]
    assert nn.sum() == 1


def test_num2numpy():
    assert num2numpy(None, 'int8') is None
    assert num2numpy(-1, 'int8').dtype == np.dtype('int8')
    assert num2numpy(-1, 'int8').dtype == np.int8(-1)

    assert num2numpy(-1, 'uint8') is None
    assert num2numpy(256, 'uint8') is None
    assert num2numpy(-1, 'uint16') is None
    assert num2numpy(-1, 'uint32') is None
    assert num2numpy(-1, 'uint8', ignore_range=True) == np.uint8(255)

    assert num2numpy(0, 'uint8') == 0
    assert num2numpy(255, 'uint8') == 255
    assert num2numpy(-128, 'int8') == -128
    assert num2numpy(127, 'int8') == 127
    assert num2numpy(128, 'int8') is None

    assert num2numpy(3.3, np.dtype('float32')).dtype == np.dtype('float32')
    assert num2numpy(3.3, np.float32).dtype == np.dtype('float32')
    assert num2numpy(3.3, np.float64).dtype == np.dtype('float64')


def test_utils_datares():
    assert data_resolution_and_offset(np.array([1.5, 2.5, 3.5])) == (1.0, 1.0)
    assert data_resolution_and_offset(np.array([5, 3, 1])) == (-2.0, 6.0)
    assert data_resolution_and_offset(np.array([5, 3])) == (-2.0, 6.0)
    assert data_resolution_and_offset(np.array([1.5]), 1) == (1.0, 1.0)

    with pytest.raises(ValueError):
        data_resolution_and_offset(np.array([]))

    with pytest.raises(ValueError):
        data_resolution_and_offset(np.array([]), 10)

    with pytest.raises(ValueError):
        data_resolution_and_offset(np.array([1]))


def test_utils_affine_from_axis():
    assert affine_from_axis(np.asarray([1.5, 2.5, 3.5]),
                            np.asarray([10.5, 11.5])) * (0, 0) == (1.0, 10.0)

    assert affine_from_axis(np.asarray([1.5, 2.5, 3.5]),
                            np.asarray([10.5, 11.5])) * (2, 1) == (3, 11)

    (sx, z1, tx,
     z2, sy, ty, *_) = affine_from_axis(np.asarray([1, 2, 3]),
                                        np.asarray([10, 20]))
    assert z1 == 0 and z2 == 0
    assert sx == 1 and sy == 10
    assert tx == 0.5 and ty == 5

    (sx, _, tx,
     _, sy, ty, *_) = affine_from_axis(np.asarray([1]),
                                       np.asarray([10, 20]), 1)
    assert sx == 1 and sy == 10
    assert tx == 0.5 and ty == 5

    (sx, _, tx,
     _, sy, ty, *_) = affine_from_axis(np.asarray([1]),
                                       np.asarray([10, 20]), (1, 1))
    assert sx == 1 and sy == 10
    assert tx == 0.5 and ty == 5

    (sx, _, tx,
     _, sy, ty, *_) = affine_from_axis(np.asarray([1]),
                                       np.asarray([10]), (2, 10))
    assert sx == 2 and sy == 10
    assert tx == 0 and ty == 5


def test_utils_math():
    xx = xr.DataArray(np.zeros((3, 4)),
                      name='xx',
                      dims=('y', 'x'),
                      coords={'x': np.arange(4),
                              'y': np.arange(3)})
    xx_t = unsqueeze_data_array(xx, 'time', 0)
    assert xx_t.dims == ('time', 'y', 'x')
    assert 'time' in xx_t.coords
    assert xx_t.data.shape == (1, 3, 4)

    ds = unsqueeze_dataset(xx.to_dataset(), 'time')
    assert ds.xx.dims == ('time', 'y', 'x')
    assert 'time' in ds.xx.coords
    assert ds.xx.data.shape == (1, 3, 4)


def test_spatial_dims():
    def tt(d1, d2):
        coords = {}
        coords[d1] = np.arange(3)
        coords[d2] = np.arange(4)
        return xr.DataArray(np.zeros((3, 4)),
                            name='xx',
                            dims=(d1, d2),
                            coords=coords)

    assert spatial_dims(tt('y', 'x')) == ('y', 'x')
    assert spatial_dims(tt('x', 'y')) == ('y', 'x')
    assert spatial_dims(tt('latitude', 'longitude')) == ('latitude', 'longitude')
    assert spatial_dims(tt('lat', 'lon')) == ('lat', 'lon')
    assert spatial_dims(tt('a', 'b')) is None
    assert spatial_dims(tt('a', 'b'), relaxed=True) == ('a', 'b')


def test_check_write_path(tmpdir):
    tmpdir = Path(str(tmpdir))
    some_path = tmpdir/"_should_not_exist-5125177.txt"
    assert not some_path.exists()
    assert check_write_path(some_path, overwrite=False) is some_path
    assert check_write_path(str(some_path), overwrite=False) == some_path
    assert isinstance(check_write_path(str(some_path), overwrite=False), Path)

    p = tmpdir/"ttt.tmp"
    with open(str(p), 'wt') as f:
        f.write("text")

    assert p.exists()
    with pytest.raises(IOError):
        check_write_path(p, overwrite=False)

    assert check_write_path(p, overwrite=True) == p
    assert not p.exists()



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
""" rasterio environment management tools
"""
import threading
from types import SimpleNamespace
import rasterio                                         # type: ignore[import]
from rasterio.session import AWSSession, DummySession   # type: ignore[import]
import rasterio.env                                     # type: ignore[import]
from datacube.utils.generic import thread_local_cache

_CFG_LOCK = threading.Lock()
_CFG = SimpleNamespace(aws=None, cloud_defaults=False, kwargs={}, epoch=0)


SECRET_KEYS = ("AWS_ACCESS_KEY_ID", "AWS_SECRET_ACCESS_KEY", "AWS_SESSION_TOKEN")


def _sanitize(opts, keys):
    return {k: (v if k not in keys else "xx..xx") for k, v in opts.items()}


def _state(purge=False):
    """
    .env   None| rasterio.Env
    .epoch -1  | +Int
    """
    return thread_local_cache(
        "__rio_state__", SimpleNamespace(env=None, epoch=-1), purge=purge
    )


def get_rio_env(sanitize=True):
    """ Get GDAL params configured by rasterio for the current thread.

    :param sanitize: If True replace sensitive Values with 'x'
    """

    env = rasterio.env.local._env  # pylint: disable=protected-access
    if env is None:
        return {}
    opts = env.get_config_options()
    if sanitize:
        opts = _sanitize(opts, SECRET_KEYS)

    return opts


def deactivate_rio_env():
    """ Exit previously configured environment, or do nothing if one wasn't configured.
    """
    state = _state(purge=True)

    if state.env is not None:
        state.env.__exit__(None, None, None)


def activate_rio_env(aws=None, cloud_defaults=False, **kwargs):
    """ Inject activated rasterio.Env into current thread.

    This de-activates previously setup environment.

    :param aws: Dictionary of options for rasterio.session.AWSSession
                OR 'auto' -- session = rasterio.session.AWSSession()

    :param cloud_defaults: When True inject settings for reading COGs
    :param **kwargs: Passed on to rasterio.Env(..) constructor
    """
    session = DummySession()

    if aws is not None:
        if not (aws == "auto" or isinstance(aws, dict)):
            raise ValueError('Only support: None|"auto"|{..} for `aws` parameter')

        aws = {} if aws == "auto" else dict(**aws)
        region_name = aws.get("region_name", "auto")

        if region_name == "auto":
            from datacube.utils.aws import auto_find_region

            try:
                aws["region_name"] = auto_find_region()
            except ValueError as e:
                # only treat it as error if it was requested by user
                if "region_name" in aws:
                    raise e

        session = AWSSession(**aws)

    opts = (
        dict(
            GDAL_DISABLE_READDIR_ON_OPEN="EMPTY_DIR",
            GDAL_HTTP_MAX_RETRY="10",
            GDAL_HTTP_RETRY_DELAY="0.5",
        )
        if cloud_defaults
        else {}
    )

    opts.update(**kwargs)

    state = _state()

    if state.env is not None:
        state.env.__exit__(None, None, None)

    env = rasterio.Env(session=session, **opts)
    env.__enter__()
    state.env = env
    state.epoch = -1

    return get_rio_env()


def activate_from_config():
    """ Check if this threads needs to reconfigure, then does reconfigure.

    - Does nothing if this thread is already configured and configuration hasn't changed.
    - Configures current thread with default rio settings
    """
    cfg = _CFG
    state = _state()

    if cfg.epoch != state.epoch:
        ee = activate_rio_env(
            aws=cfg.aws, cloud_defaults=cfg.cloud_defaults, **cfg.kwargs
        )
        state.epoch = cfg.epoch
        return ee

    return None


def set_default_rio_config(aws=None, cloud_defaults=False, **kwargs):
    """ Setup default configuration for rasterio/GDAL.

    Doesn't actually activate one, just stores configuration for future
    use from IO threads.

    :param aws: Dictionary of options for rasterio.session.AWSSession
                OR 'auto' -- session = rasterio.session.AWSSession()

    :param cloud_defaults: When True inject settings for reading COGs
    :param **kwargs: Passed on to rasterio.Env(..) constructor
    """
    global _CFG  # pylint: disable=global-statement

    with _CFG_LOCK:
        _CFG = SimpleNamespace(
            aws=aws, cloud_defaults=cloud_defaults, kwargs=kwargs, epoch=_CFG.epoch + 1
        )


def configure_s3_access(
    profile=None,
    region_name="auto",
    aws_unsigned=False,
    requester_pays=False,
    cloud_defaults=True,
    client=None,
    **gdal_opts
):
    """
    Use :meth:`datacube.utils.aws.configure_s3_access` instead.
    """
    from datacube.utils.aws import configure_s3_access

    return configure_s3_access(
        profile=profile,
        region_name=region_name,
        aws_unsigned=aws_unsigned,
        requester_pays=requester_pays,
        cloud_defaults=cloud_defaults,
        client=client,
        **gdal_opts
    )



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import pytest
from datacube.testutils.threads import FakeThreadPoolExecutor
from datacube.testutils import mk_sample_xr_dataset, mk_sample_dataset
from datacube.testutils.io import native_geobox


def test_fakethreadpool():

    def tfunc(a: int, b: int = 0, please_fail=False) -> int:
        if please_fail:
            raise ValueError('as you wish')
        if a == 13:
            raise ValueError('13')
        return a + b

    pool = FakeThreadPoolExecutor()

    assert pool.submit(tfunc, 1).result() == 1
    assert pool.submit(tfunc, 1, 2).result() == 3

    fut = pool.submit(tfunc, 1, please_fail=True)
    assert fut.done()
    assert fut.exception() is not None

    with pytest.raises(ValueError):
        fut.result()

    ff = list(pool.map(tfunc, range(14)))
    assert len(ff) == 14
    assert [f.result() for f in ff[:13]] == list(range(13))
    assert ff[13].exception() is not None

    aa = list(range(10))
    bb = aa[::-1]
    ff = list(pool.map(tfunc, aa, bb))
    assert len(ff) == 10
    assert [f.result() for f in ff[:13]] == [a+b for a, b in zip(aa, bb)]

    pool.shutdown()


def test_mk_sample_xr():
    xx = mk_sample_xr_dataset()
    assert 'band' in xx.data_vars
    assert list(xx.coords) == ['time', 'y', 'x', 'spatial_ref']
    assert xx.band.dims == ('time', 'y', 'x')
    assert xx.geobox is not None

    assert mk_sample_xr_dataset(name='xx', shape=(3, 7)).xx.shape == (1, 3, 7)
    assert mk_sample_xr_dataset(name='xx', time=None, shape=(3, 7)).xx.shape == (3, 7)
    assert mk_sample_xr_dataset(name='xx', time=None).xx.dims == ('y', 'x')

    assert mk_sample_xr_dataset(resolution=(1, 100)).geobox.resolution == (1, 100)
    assert mk_sample_xr_dataset(resolution=(1, 100), xy=(3, 55)).geobox.transform*(0, 0) == (3, 55)
    assert mk_sample_xr_dataset(crs=None).geobox is None


def test_native_geobox_ingested():
    from datacube.testutils.io import native_geobox
    from datacube.testutils.geom import AlbersGS

    gbox = AlbersGS.tile_geobox((15, -40))
    ds = mk_sample_dataset([dict(name='a')],
                           geobox=gbox,
                           product_opts=dict(with_grid_spec=True))

    assert native_geobox(ds) == gbox

    # check that dataset covering several tiles is detected as invalid
    ds = mk_sample_dataset([dict(name='a')],
                           geobox=gbox.buffered(10, 10),
                           product_opts=dict(with_grid_spec=True))

    with pytest.raises(ValueError):
        native_geobox(ds)


def test_native_geobox_eo3(eo3_dataset_s2):
    ds = eo3_dataset_s2
    assert ds.crs is not None
    assert 'blue' in ds.measurements

    gb1 = native_geobox(ds, basis='blue')

    assert gb1.width == 10980
    assert gb1.height == 10980
    assert gb1.crs == ds.crs

    gb1_ = native_geobox(ds, ('blue', 'red', 'green'))
    assert gb1_ == gb1

    gb2 = native_geobox(ds, ['swir_1', 'swir_2'])
    assert gb2 != gb1
    assert gb2.width == 5490
    assert gb2.height == 5490
    assert gb2.crs == ds.crs

    with pytest.raises(ValueError):
        native_geobox(ds)

    with pytest.raises(ValueError):
        native_geobox(ds, ['no_such_band'])

    ds.metadata_doc['measurements']['red_edge_1']['grid'] = 'no-such-grid'

    with pytest.raises(ValueError):
        native_geobox(ds, ['red_edge_1'])



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
from setuptools import setup, find_packages

setup(
    name='dcio_example',
    version='1.1',
    description='Test IO plugins for datacube',
    author='Open Data Cube',
    packages=find_packages(),

    entry_points={
        'datacube.plugins.io.read': [
            'pickle=dcio_example.pickles:rdr_driver_init',
            'zeros=dcio_example.zeros:init_driver',
            'xarray_3d=dcio_example.xarray_3d:reader_driver_init',
        ],
        'datacube.plugins.io.write': [
            'pickle=dcio_example.pickles:writer_driver_init',
        ],
    },
)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import os
from pathlib import Path
from typing import Union, Optional


def _norm_path(path: Union[str, Path], in_home_dir: bool = False) -> Path:
    if isinstance(path, str):
        path = Path(path)
    if in_home_dir:
        path = Path.home()/path
    return path


def check_write_path(fname: Union[Path, str], overwrite: bool) -> Path:
    """ Check is output file exists and either remove it first or raise IOError.

    :param fname: string or Path object
    :param overwrite: Whether to remove file when it exists

    exists   overwrite   Action
    ----------------------------------------------
    T            T       delete file, return Path
    T            F       raise IOError
    F            T       return Path
    F            F       return Path
    """
    if not isinstance(fname, Path):
        fname = Path(fname)

    if fname.exists():
        if overwrite:
            fname.unlink()
        else:
            raise IOError("File exists")
    return fname


def write_user_secret_file(text: Union[str, bytes],
                           fname: Union[str, Path],
                           in_home_dir: bool = False,
                           mode: str = 'w'):
    """Write file only readable/writeable by the user"""

    fname = _norm_path(fname, in_home_dir)
    open_flags = os.O_WRONLY | os.O_CREAT | os.O_TRUNC
    access = 0o600  # Make sure file is readable by current user only
    with os.fdopen(os.open(str(fname), open_flags, access), mode) as handle:
        handle.write(text)
        handle.close()


def slurp(fname: Union[str, Path],
          in_home_dir: bool = False,
          mode: str = 'r') -> Optional[Union[bytes, str]]:
    """
    Read an entire file into a string

    :param fname: file path
    :param in_home_dir: if True treat fname as a path relative to $HOME folder
    :return: Content of a file or None if file doesn't exist or can not be read for any other reason
    """
    fname = _norm_path(fname, in_home_dir)
    try:
        with open(str(fname), mode) as handle:
            return handle.read()
    except IOError:
        return None



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Important functions are:

* :func:`reproject_and_fuse`

"""
import logging
from collections import OrderedDict
import numpy as np
from xarray.core.dataarray import DataArray as XrDataArray, DataArrayCoordinates
from xarray.core.dataset import Dataset as XrDataset
from typing import (
    Union, Optional, Callable,
    List, Any, Iterator, Iterable, Mapping, Tuple, Hashable, cast
)

from datacube.utils import ignore_exceptions_if
from datacube.utils.math import invalid_mask
from datacube.utils.geometry import GeoBox, roi_is_empty
from datacube.model import Measurement
from datacube.drivers._types import ReaderDriver
from ..drivers.datasource import DataSource
from ._base import BandInfo

_LOG = logging.getLogger(__name__)

FuserFunction = Callable[[np.ndarray, np.ndarray], Any]  # pylint: disable=invalid-name
ProgressFunction = Callable[[int, int], Any]  # pylint: disable=invalid-name


def _default_fuser(dst: np.ndarray, src: np.ndarray, dst_nodata) -> None:
    """ Overwrite only those pixels in `dst` with `src` that are "not valid"

        For every pixel in dst that equals to dst_nodata replace it with pixel
        from src.
    """
    np.copyto(dst, src, where=invalid_mask(dst, dst_nodata))


def reproject_and_fuse(datasources: List[DataSource],
                       destination: np.ndarray,
                       dst_gbox: GeoBox,
                       dst_nodata: Optional[Union[int, float]],
                       resampling: str = 'nearest',
                       fuse_func: Optional[FuserFunction] = None,
                       skip_broken_datasets: bool = False,
                       progress_cbk: Optional[ProgressFunction] = None,
                       extra_dim_index: Optional[int] = None):
    """
    Reproject and fuse `sources` into a 2D numpy array `destination`.

    :param datasources: Data sources to open and read from
    :param destination: ndarray of appropriate size to read data into
    :param dst_gbox: GeoBox defining destination region
    :param skip_broken_datasets: Carry on in the face of adversity and failing reads.
    :param progress_cbk: If supplied will be called with 2 integers `Items processed, Total Items`
                         after reading each file.
    """
    # pylint: disable=too-many-locals
    from ._read import read_time_slice
    assert len(destination.shape) == 2

    def copyto_fuser(dest: np.ndarray, src: np.ndarray) -> None:
        _default_fuser(dest, src, dst_nodata)

    fuse_func = fuse_func or copyto_fuser

    destination.fill(dst_nodata)
    if len(datasources) == 0:
        return destination
    elif len(datasources) == 1:
        with ignore_exceptions_if(skip_broken_datasets):
            with datasources[0].open() as rdr:
                read_time_slice(rdr, destination, dst_gbox, resampling, dst_nodata, extra_dim_index)

        if progress_cbk:
            progress_cbk(1, 1)

        return destination
    else:
        # Multiple sources, we need to fuse them together into a single array
        buffer_ = np.full(destination.shape, dst_nodata, dtype=destination.dtype)
        for n_so_far, source in enumerate(datasources, 1):
            with ignore_exceptions_if(skip_broken_datasets):
                with source.open() as rdr:
                    roi = read_time_slice(rdr, buffer_, dst_gbox, resampling, dst_nodata, extra_dim_index)

                if not roi_is_empty(roi):
                    fuse_func(destination[roi], buffer_[roi])
                    buffer_[roi] = dst_nodata  # clean up for next read

            if progress_cbk:
                progress_cbk(n_so_far, len(datasources))

        return destination


def _mk_empty_ds(coords: DataArrayCoordinates, geobox: GeoBox) -> XrDataset:
    cc = OrderedDict(coords.items())
    cc.update(geobox.xr_coords())
    return XrDataset(coords=cast(Mapping[Hashable, Any], cc), attrs={'crs': geobox.crs})


def _allocate_storage(coords: DataArrayCoordinates,
                      geobox: GeoBox,
                      measurements: Iterable[Measurement]) -> XrDataset:
    xx = _mk_empty_ds(coords, geobox)
    dims = list(xx.coords.keys())
    shape = tuple(xx.sizes[k] for k in dims)

    for m in measurements:
        name, dtype, attrs = m.name, m.dtype, m.dataarray_attrs()
        attrs['crs'] = geobox.crs
        data = np.empty(shape, dtype=dtype)
        xx[name] = XrDataArray(data, coords=xx.coords, dims=dims, name=name, attrs=attrs)

    return xx


def xr_load(sources: XrDataArray,
            geobox: GeoBox,
            measurements: List[Measurement],
            driver: ReaderDriver,
            driver_ctx_prev: Optional[Any] = None,
            skip_broken_datasets: bool = False) -> Tuple[XrDataset, Any]:
    # pylint: disable=too-many-locals
    from ._read import read_time_slice_v2

    out = _allocate_storage(sources.coords, geobox, measurements)

    def all_groups() -> Iterator[Tuple[Measurement, Tuple[int, ...], List[BandInfo]]]:
        for idx, dss in np.ndenumerate(sources.values):
            for m in measurements:
                bbi = [BandInfo(ds, m.name) for ds in dss]
                yield (m, idx, bbi)

    def just_bands(groups) -> Iterator[BandInfo]:
        for _, _, bbi in groups:
            yield from bbi

    groups = list(all_groups())
    ctx = driver.new_load_context(just_bands(groups), driver_ctx_prev)

    # TODO: run upto N concurrently
    for m, idx, bbi in groups:
        dst = out.data_vars[m.name].values[idx]
        dst[:] = m.nodata
        resampling = m.get('resampling_method', 'nearest')
        fuse_func = m.get('fuser', None)

        for band in bbi:
            rdr = driver.open(band, ctx).result()

            pix, roi = read_time_slice_v2(rdr, geobox, resampling, m.nodata)

            if pix is not None:
                if fuse_func:
                    fuse_func(dst[roi], pix)
                else:
                    _default_fuser(dst[roi], pix, m.nodata)

    return out, ctx



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import netCDF4
import numpy
import xarray as xr
import pytest
from hypothesis import given
from hypothesis.strategies import text
import string
from uuid import uuid4

from datacube.drivers.netcdf._write import _get_units
from datacube.drivers.netcdf.writer import create_netcdf, create_coordinate, create_variable, netcdfy_data, \
    create_grid_mapping_variable, flag_mask_meanings, Variable
from datacube.drivers.netcdf import write_dataset_to_netcdf
from datacube.drivers.netcdf.writer import DEFAULT_GRID_MAPPING
from datacube.utils import geometry, DatacubeException, read_strings_from_netcdf
from datacube.testutils import mk_sample_xr_dataset


crs_var = DEFAULT_GRID_MAPPING

GEO_PROJ = geometry.CRS("""GEOGCS["WGS 84",
                           DATUM["WGS_1984",SPHEROID["WGS 84",6378137,298.257223563,AUTHORITY["EPSG","7030"]],
                           AUTHORITY["EPSG","6326"]],PRIMEM["Greenwich",0],UNIT["degree",0.0174532925199433],
                           AUTHORITY["EPSG","4326"]]""")

ALBERS_PROJ = geometry.CRS("""PROJCS["GDA94 / Australian Albers",
                                GEOGCS["GDA94",
                                    DATUM["Geocentric_Datum_of_Australia_1994",
                                        SPHEROID["GRS 1980",6378137,298.257222101,
                                            AUTHORITY["EPSG","7019"]],
                                        TOWGS84[0,0,0,0,0,0,0],
                                        AUTHORITY["EPSG","6283"]],
                                    PRIMEM["Greenwich",0,
                                        AUTHORITY["EPSG","8901"]],
                                    UNIT["degree",0.01745329251994328,
                                        AUTHORITY["EPSG","9122"]],
                                    AUTHORITY["EPSG","4283"]],
                                UNIT["metre",1,
                                    AUTHORITY["EPSG","9001"]],
                                PROJECTION["Albers_Conic_Equal_Area"],
                                PARAMETER["standard_parallel_1",-18],
                                PARAMETER["standard_parallel_2",-36],
                                PARAMETER["latitude_of_center",0],
                                PARAMETER["longitude_of_center",132],
                                PARAMETER["false_easting",0],
                                PARAMETER["false_northing",0],
                                AUTHORITY["EPSG","3577"],
                                AXIS["Easting",EAST],
                                AXIS["Northing",NORTH]]""")

SINIS_PROJ = geometry.CRS("""PROJCS["Sinusoidal",
                                GEOGCS["GCS_Undefined",
                                        DATUM["Undefined",
                                        SPHEROID["User_Defined_Spheroid",6371007.181,0.0]],
                                        PRIMEM["Greenwich",0.0],
                                        UNIT["Degree",0.0174532925199433]],
                                PROJECTION["Sinusoidal"],
                                PARAMETER["False_Easting",0.0],
                                PARAMETER["False_Northing",0.0],
                                PARAMETER["Central_Meridian",0.0],
                                UNIT["Meter",1.0]]""")

LCC2_PROJ = geometry.CRS("""PROJCS["unnamed",
                               GEOGCS["WGS 84",
                                       DATUM["unknown",
                                       SPHEROID["WGS84",6378137,6556752.3141]],
                                       PRIMEM["Greenwich",0],
                                       UNIT["degree",0.0174532925199433]],
                               UNIT["metre",1, AUTHORITY["EPSG","9001"]],
                               PROJECTION["Lambert_Conformal_Conic_2SP"],
                               PARAMETER["standard_parallel_1",17.5],
                               PARAMETER["standard_parallel_2",29.5],
                               PARAMETER["latitude_of_origin",12],
                               PARAMETER["central_meridian",-102],
                               PARAMETER["false_easting",2500000],
                               PARAMETER["false_northing",0]]""")


def _ensure_spheroid(var):
    assert 'semi_major_axis' in var.ncattrs()
    assert 'semi_minor_axis' in var.ncattrs()
    assert 'inverse_flattening' in var.ncattrs()


def _ensure_gdal(var):
    assert 'GeoTransform' in var.ncattrs()
    assert 'spatial_ref' in var.ncattrs()


def _ensure_geospatial(nco):
    assert 'geospatial_bounds' in nco.ncattrs()
    assert 'geospatial_bounds_crs' in nco.ncattrs()
    assert nco.getncattr('geospatial_bounds_crs') == "EPSG:4326"

    assert 'geospatial_lat_min' in nco.ncattrs()
    assert 'geospatial_lat_max' in nco.ncattrs()
    assert 'geospatial_lat_units' in nco.ncattrs()
    assert nco.getncattr('geospatial_lat_units') == "degrees_north"

    assert 'geospatial_lon_min' in nco.ncattrs()
    assert 'geospatial_lon_max' in nco.ncattrs()
    assert 'geospatial_lon_units' in nco.ncattrs()
    assert nco.getncattr('geospatial_lon_units') == "degrees_east"


def test_create_albers_projection_netcdf(tmpnetcdf_filename):
    nco = create_netcdf(tmpnetcdf_filename)
    create_coordinate(nco, 'x', numpy.array([1., 2., 3.]), 'm')
    create_coordinate(nco, 'y', numpy.array([1., 2., 3.]), 'm')
    create_grid_mapping_variable(nco, ALBERS_PROJ)
    nco.close()

    with netCDF4.Dataset(tmpnetcdf_filename) as nco:
        assert crs_var in nco.variables
        assert nco[crs_var].grid_mapping_name == 'albers_conical_equal_area'
        assert 'standard_parallel' in nco[crs_var].ncattrs()
        assert 'longitude_of_central_meridian' in nco[crs_var].ncattrs()
        assert 'latitude_of_projection_origin' in nco[crs_var].ncattrs()
        _ensure_spheroid(nco[crs_var])
        _ensure_gdal(nco[crs_var])
        _ensure_geospatial(nco)


def test_create_lambert_conformal_conic_2sp_projection_netcdf(tmpnetcdf_filename):
    nco = create_netcdf(tmpnetcdf_filename)
    create_coordinate(nco, 'x', numpy.array([1., 2., 3.]), 'm')
    create_coordinate(nco, 'y', numpy.array([1., 2., 3.]), 'm')
    create_grid_mapping_variable(nco, LCC2_PROJ)
    nco.close()

    with netCDF4.Dataset(tmpnetcdf_filename) as nco:
        assert crs_var in nco.variables
        assert nco[crs_var].grid_mapping_name == 'lambert_conformal_conic'
        assert 'standard_parallel' in nco[crs_var].ncattrs()
        assert 'longitude_of_central_meridian' in nco[crs_var].ncattrs()
        assert 'latitude_of_projection_origin' in nco[crs_var].ncattrs()
        assert 'false_easting' in nco[crs_var].ncattrs()
        assert 'false_northing' in nco[crs_var].ncattrs()
        _ensure_spheroid(nco[crs_var])
        _ensure_gdal(nco[crs_var])
        _ensure_geospatial(nco)


def test_create_epsg4326_netcdf(tmpnetcdf_filename):
    nco = create_netcdf(tmpnetcdf_filename)
    create_coordinate(nco, 'latitude', numpy.array([1., 2., 3.]), 'm')
    create_coordinate(nco, 'longitude', numpy.array([1., 2., 3.]), 'm')
    create_grid_mapping_variable(nco, GEO_PROJ)
    nco.close()

    with netCDF4.Dataset(tmpnetcdf_filename) as nco:
        assert crs_var in nco.variables
        assert nco[crs_var].grid_mapping_name == 'latitude_longitude'
        _ensure_spheroid(nco[crs_var])
        _ensure_geospatial(nco)


def test_create_sinus_netcdf(tmpnetcdf_filename):
    nco = create_netcdf(tmpnetcdf_filename)
    create_coordinate(nco, 'x', numpy.array([1., 2., 3.]), 'm')
    create_coordinate(nco, 'y', numpy.array([1., 2., 3.]), 'm')
    create_grid_mapping_variable(nco, SINIS_PROJ)
    nco.close()

    with netCDF4.Dataset(tmpnetcdf_filename) as nco:
        assert crs_var in nco.variables
        assert nco[crs_var].grid_mapping_name == 'sinusoidal'
        assert 'longitude_of_central_meridian' in nco[crs_var].ncattrs()
        _ensure_spheroid(nco[crs_var])
        _ensure_geospatial(nco)


# Work around outstanding bug with hypothesis/pytest, where function level fixtures are only run once.
# Generate a new netcdf filename for each run, so that old files don't cause permission errors on windows
# due to antivirus software filesystem lag.
# See https://github.com/HypothesisWorks/hypothesis-python/issues/377
@given(s1=text(alphabet=string.printable, max_size=100),
       s2=text(alphabet=string.printable, max_size=100),
       s3=text(alphabet=string.printable, max_size=100))
def test_create_string_variable(workdir, s1, s2, s3):
    tmpnetcdf_filename = workdir.join(f"testfile_np_{uuid4()}.nc")
    str_var = 'str_var'
    nco = create_netcdf(tmpnetcdf_filename)
    coord = create_coordinate(nco, 'greg', numpy.array([1.0, 3.0, 9.0]), 'cubic gregs')
    assert coord is not None

    dtype = numpy.dtype('S100')
    data = numpy.array([s1, s2, s3], dtype=dtype)

    var = create_variable(nco, str_var, Variable(dtype, None, ('greg',), None))
    var[:] = netcdfy_data(data)
    nco.close()

    with netCDF4.Dataset(tmpnetcdf_filename) as nco:
        assert str_var in nco.variables

    for returned, expected in zip(read_strings_from_netcdf(tmpnetcdf_filename, variable=str_var), (s1, s2, s3)):
        assert returned == expected


def test_chunksizes(tmpnetcdf_filename):
    nco = create_netcdf(tmpnetcdf_filename)

    x = numpy.arange(3, dtype='float32')
    y = numpy.arange(5, dtype='float32')

    coord1 = create_coordinate(nco, 'x', x, 'm')
    coord2 = create_coordinate(nco, 'y', y, 'm')

    assert coord1 is not None and coord2 is not None

    no_chunks = create_variable(nco, 'no_chunks',
                                Variable(numpy.dtype('int16'), None, ('x', 'y'), None))

    min_max_chunks = create_variable(nco, 'min_max_chunks',
                                     Variable(numpy.dtype('int16'), None, ('x', 'y'), None),
                                     chunksizes=(2, 50))

    assert no_chunks is not None
    assert min_max_chunks is not None

    strings = numpy.array(["AAa", 'bbb', 'CcC'], dtype='S')
    strings = xr.DataArray(strings, dims=['x'], coords={'x': x})
    create_variable(nco, 'strings_unchunked', strings)
    create_variable(nco, 'strings_chunked', strings, chunksizes=(1,))

    nco.close()

    with netCDF4.Dataset(tmpnetcdf_filename) as nco:
        assert nco['no_chunks'].chunking() == 'contiguous'
        assert nco['min_max_chunks'].chunking() == [2, 5]
        assert nco['strings_unchunked'].chunking() == 'contiguous'
        assert nco['strings_chunked'].chunking() == [1, 3]


EXAMPLE_FLAGS_DEF = {
    'band_1_saturated': {
        'bits': 0,
        'values': {
            0: True,
            1: False
        },
        'description': 'Band 1 is saturated'
    },
    'band_2_saturated': {
        'bits': 1,
        'values': {
            0: True,
            1: False
        },
        'description': 'Band 2 is saturated'
    },
    'band_3_saturated': {
        'bits': 2,
        'values': {
            0: True,
            1: False
        },
        'description': 'Band 3 is saturated'
    },
    'land_sea': {
        'bits': 9,
        'values': {
            0: 'sea',
            1: 'land'
        },
        'description': 'Land/Sea observation'
    },
}


def test_measurements_model_netcdfflags():
    masks, valid_range, meanings = flag_mask_meanings(EXAMPLE_FLAGS_DEF)
    assert ([0, 1023] == valid_range).all()
    assert ([1, 2, 4, 512] == masks).all()
    assert 'no_band_1_saturated no_band_2_saturated no_band_3_saturated land' == meanings


def test_useful_error_on_write_empty_dataset(tmpnetcdf_filename):
    with pytest.raises(DatacubeException) as excinfo:
        ds = xr.Dataset()
        write_dataset_to_netcdf(ds, tmpnetcdf_filename)
    assert 'empty' in str(excinfo.value)

    with pytest.raises(DatacubeException) as excinfo:
        ds = xr.Dataset(data_vars={'blue': (('time',), numpy.array([0, 1, 2]))})
        write_dataset_to_netcdf(ds, tmpnetcdf_filename)
    assert 'geobox' in str(excinfo.value)


def test_write_dataset_to_netcdf(tmpnetcdf_filename, odc_style_xr_dataset):
    write_dataset_to_netcdf(odc_style_xr_dataset, tmpnetcdf_filename, global_attributes={'foo': 'bar'},
                            variable_params={'B10': {'attrs': {'abc': 'xyz'}}})

    with netCDF4.Dataset(tmpnetcdf_filename) as nco:
        nco.set_auto_mask(False)
        assert 'B10' in nco.variables
        var = nco.variables['B10']
        assert (var[:] == odc_style_xr_dataset['B10'].values).all()

        assert 'foo' in nco.ncattrs()
        assert nco.getncattr('foo') == 'bar'

        assert 'abc' in var.ncattrs()
        assert var.getncattr('abc') == 'xyz'

    with pytest.raises(RuntimeError):
        write_dataset_to_netcdf(odc_style_xr_dataset, tmpnetcdf_filename)

    # Check grid_mapping is a coordinate
    xx = xr.open_dataset(tmpnetcdf_filename)
    assert crs_var in xx.coords
    assert crs_var not in xx.data_vars


def test_write_dataset_with_time_dimension_to_netcdf(tmpnetcdf_filename):
    xx = mk_sample_xr_dataset(name='B10', time='2020-01-01')
    assert 'time' in xx.coords
    assert 'units' not in xx.time.attrs

    write_dataset_to_netcdf(xx, tmpnetcdf_filename, global_attributes={'foo': 'bar'},
                            variable_params={'B10': {'attrs': {'abc': 'xyz'}}})

    with netCDF4.Dataset(tmpnetcdf_filename) as nco:
        nco.set_auto_mask(False)
        assert 'B10' in nco.variables
        var = nco.variables['B10']
        assert (var[:] == xx['B10'].values).all()

        assert 'foo' in nco.ncattrs()
        assert nco.getncattr('foo') == 'bar'

        assert 'abc' in var.ncattrs()
        assert var.getncattr('abc') == 'xyz'

    with pytest.raises(RuntimeError):
        write_dataset_to_netcdf(xx, tmpnetcdf_filename)

    # Check grid_mapping is a coordinate
    yy = xr.open_dataset(tmpnetcdf_filename)
    assert crs_var in yy.coords
    assert crs_var not in yy.data_vars

    assert 'time' in yy.coords


def test_get_units():
    assert _get_units(geometry.Coordinate([], 'K', None)) == 'K'
    assert _get_units(geometry.Coordinate([], None, None)) == '1'
    assert _get_units(geometry.Coordinate(numpy.zeros(1, dtype='uint8'), None, None)) == '1'
    assert _get_units(geometry.Coordinate(
        numpy.zeros(1, dtype='datetime64[s]'), None, None)).startswith('seconds since ')



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
User configuration.
"""

import os
from pathlib import Path
import configparser
from urllib.parse import unquote_plus, urlparse, parse_qsl
from typing import Any, Dict, Iterable, MutableMapping, Optional, Tuple, Union

PathLike = Union[str, 'os.PathLike[Any]']

ConfigDict = Dict[str, Union[str, int, bool]]

ENVIRONMENT_VARNAME = 'DATACUBE_CONFIG_PATH'
#: Config locations in order. Properties found in latter locations override
#: earlier ones.
#:
#: - `/etc/datacube.conf`
#: - file at `$DATACUBE_CONFIG_PATH` environment variable
#: - `~/.datacube.conf`
#: - `datacube.conf`
DEFAULT_CONF_PATHS = tuple(p for p in ['/etc/datacube.conf',
                                       os.environ.get(ENVIRONMENT_VARNAME, ''),
                                       str(os.path.expanduser("~/.datacube.conf")),
                                       'datacube.conf'] if len(p) > 0)

DEFAULT_ENV = 'default'

# Default configuration options.
_DEFAULT_CONF = """
[DEFAULT]
# Blank implies localhost
db_hostname:
db_database: datacube
index_driver: default
# If a connection is unused for this length of time, expect it to be invalidated.
db_connection_timeout: 60

[user]
# Which environment to use when none is specified explicitly.
#   note: will fail if default_environment points to non-existent section
# default_environment: datacube
"""

#: Used in place of None as a default, when None is a valid but not default parameter to a function
_UNSET = object()


def read_config(default_text: Optional[str] = None) -> configparser.ConfigParser:
    config = configparser.ConfigParser()
    if default_text is not None:
        config.read_string(default_text)
    return config


class LocalConfig(object):
    """
    System configuration for the user.

    This loads from a set of possible configuration files which define the available environments.
    An environment contains connection details for a Data Cube Index, which provides access to
    available data.

    """

    def __init__(self, config: configparser.ConfigParser,
                 files_loaded: Optional[Iterable[str]] = None,
                 env: Optional[str] = None):
        """
        Datacube environment resolution precedence is:
          1. Supplied as a function argument `env`
          2. DATACUBE_ENVIRONMENT environment variable
          3. user.default_environment option in the config
          4. 'default' or 'datacube' whichever is present

        If environment is supplied by any of the first 3 methods is not present
        in the config, then throw an exception.
        """
        self._config = config
        self.files_loaded = [] if files_loaded is None else list(iter(files_loaded))

        if env is None:
            env = os.environ.get('DATACUBE_ENVIRONMENT',
                                 config.get('user', 'default_environment', fallback=None))

        # If the user specifies a particular env, we either want to use it or Fail
        if env:
            if config.has_section(env):
                self._env = env
                # All is good
                return
            else:
                raise ValueError('No config section found for environment %r' % (env,))
        else:
            # If an env hasn't been specifically selected, we can fall back defaults
            fallbacks = [DEFAULT_ENV, 'datacube']
            for fallback_env in fallbacks:
                if config.has_section(fallback_env):
                    self._env = fallback_env
                    return
            raise ValueError('No ODC environment, checked configurations for %s' % fallbacks)

    @classmethod
    def find(cls,
             paths: Optional[Union[str, Iterable[PathLike]]] = None,
             env: Optional[str] = None) -> 'LocalConfig':
        """
        Find config from environment variables or possible filesystem locations.

        'env' is which environment to use from the config: it corresponds to the name of a
        config section
        """
        config = read_config(_DEFAULT_CONF)

        if paths is None:
            if env is None:
                env_opts = parse_env_params()
                if env_opts:
                    return _cfg_from_env_opts(env_opts, config)

            paths = DEFAULT_CONF_PATHS

        if isinstance(paths, str) or hasattr(paths, '__fspath__'):  # Use os.PathLike in 3.6+
            paths = [str(paths)]

        files_loaded = config.read(str(p) for p in paths if p)

        return LocalConfig(
            config,
            files_loaded=files_loaded,
            env=env,
        )

    def get(self, item: str, fallback=_UNSET):
        if fallback is _UNSET:
            return self._config.get(self._env, item)
        else:
            return self._config.get(self._env, item, fallback=fallback)

    def __getitem__(self, item: str):
        return self.get(item, fallback=None)

    def __str__(self) -> str:
        cfg = dict(self._config[self._env])
        if 'db_password' in cfg:
            cfg['db_password'] = '***'

        return "LocalConfig<loaded_from={}, environment={!r}, config={}>".format(
            self.files_loaded or 'defaults',
            self._env,
            cfg)

    def __repr__(self) -> str:
        return str(self)


DB_KEYS = ('hostname', 'port', 'database', 'username', 'password')


def parse_connect_url(url: str) -> ConfigDict:
    """ Extract database,hostname,port,username,password from db URL.

    Example: postgresql://username:password@hostname:port/database

    For local password-less db use `postgresql:///<your db>`
    """
    def split2(s: str, separator: str) -> Tuple[str, str]:
        i = s.find(separator)
        return (s, '') if i < 0 else (s[:i], s[i+1:])

    _, netloc, path, _, query, *_ = urlparse(url)

    db = path[1:] if path else ''
    if '@' in netloc:
        (user, password), (host, port) = (split2(p, ':') for p in split2(netloc, '@'))
    else:
        user, password = '', ''
        host, port = split2(netloc, ':')

    oo: ConfigDict = dict(hostname=host, database=db)

    if port:
        oo['port'] = port
    if password:
        oo['password'] = unquote_plus(password)
    if user:
        oo['username'] = user

    supported_keys = {
        'user': 'username',
        'host': 'hostname',
        'dbname': 'database',
        'password': 'password',
        'port': 'port',
    }
    oo.update({supported_keys[k]: v for k, v in parse_qsl(query) if k in supported_keys})

    return oo


def parse_env_params() -> ConfigDict:
    """
    - Read DATACUBE_IAM_* environment variables.
    - Extract parameters from DATACUBE_DB_URL if present
    - Else look for DB_HOSTNAME, DB_USERNAME, DB_PASSWORD, DB_DATABASE
    - Return {} otherwise
    """
    # Handle environment vars that cannot fit in the DB URL
    non_url_params: MutableMapping[str, Union[bool, int]] = {}
    iam_auth = os.environ.get('DATACUBE_IAM_AUTHENTICATION')
    if iam_auth is not None and iam_auth.lower() in ['y', 'yes']:
        non_url_params["iam_authentication"] = True
        iam_auth_timeout = os.environ.get('DATACUBE_IAM_TIMEOUT')
        if iam_auth_timeout:
            non_url_params["iam_timeout"] = int(iam_auth_timeout)

    # Handle environment vars that may fit in the DB URL
    db_url = os.environ.get('DATACUBE_DB_URL', None)
    if db_url is not None:
        params = parse_connect_url(db_url)
    else:
        raw_params = {k: os.environ.get('DB_{}'.format(k.upper()), None)
                      for k in DB_KEYS}
        params = {k: v
                  for k, v in raw_params.items()
                  if v is not None and v != ""}
    params.update(non_url_params)
    return params


def _cfg_from_env_opts(opts: ConfigDict,
                       base: configparser.ConfigParser) -> LocalConfig:
    def stringify(vin: Union[int, str, bool]) -> str:
        if isinstance(vin, bool):
            if vin:
                return "yes"
            else:
                return "no"
        else:
            return str(vin)
    base['default'] = {'db_'+k: stringify(v) for k, v in opts.items()}
    return LocalConfig(base, files_loaded=[], env='default')


def render_dc_config(params: ConfigDict,
                     section_name: str = 'default') -> str:
    """ Render output of parse_env_params to a string that can be written to config file.
    """
    oo = '[{}]\n'.format(section_name)
    for k in DB_KEYS:
        v = params.get(k, None)
        if v is not None:
            oo += 'db_{k}: {v}\n'.format(k=k, v=v)
    return oo


def auto_config() -> str:
    """
    Render config to $DATACUBE_CONFIG_PATH or ~/.datacube.conf, but only if doesn't exist.

    option1:
      DATACUBE_DB_URL  postgresql://user:password@host:port/database

    option2:
      DB_{HOSTNAME|PORT|USERNAME|PASSWORD|DATABASE}

    option3:
       default config
    """
    cfg_path: Optional[PathLike] = os.environ.get('DATACUBE_CONFIG_PATH', None)
    cfg_path = Path(cfg_path) if cfg_path else Path.home()/'.datacube.conf'

    if cfg_path.exists():
        return str(cfg_path)

    opts = parse_env_params()

    if len(opts) == 0:
        opts['hostname'] = ''
        opts['database'] = 'datacube'

    cfg_text = render_dc_config(opts)
    with open(str(cfg_path), 'wt') as f:
        f.write(cfg_text)

    return str(cfg_path)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
from threading import Lock
from typing import Any

DRIVER_CACHE_LOCK = Lock()
NOT_SET_MARKER = object()


def singleton_setup(obj: object,
                    key: str,
                    factory,
                    *args,
                    **kwargs) -> Any:
    """
    Does:
      obj.key = factory(*args, **kwargs)  # but only once and in a thread safe manner
      return obj.key
    """
    v = getattr(obj, key, NOT_SET_MARKER)
    if v is not NOT_SET_MARKER:
        return v

    with DRIVER_CACHE_LOCK:
        v = getattr(obj, key, NOT_SET_MARKER)
        if v is not NOT_SET_MARKER:
            # very rare multi-thread only event.
            # Other thread did it first but after this thread's initial check above.
            # Disable test cover
            return v  # pragma: no cover
        v = factory(*args, **kwargs)
        setattr(obj, key, v)
        return v



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import datetime
from collections import namedtuple
from contextlib import contextmanager
from copy import deepcopy

from uuid import UUID

from datacube.index.postgres._datasets import DatasetResource
from datacube.index.exceptions import DuplicateRecordError
from datacube.model import DatasetType, MetadataType, Dataset

_nbar_uuid = UUID('f2f12372-8366-11e5-817e-1040f381a756')
_ortho_uuid = UUID('5cf41d98-eda9-11e4-8a8e-1040f381a756')
_telemetry_uuid = UUID('4ec8fe97-e8b9-11e4-87ff-1040f381a756')

# An NBAR with source datasets. Many fields have been removed to keep it semi-focused to our ingest test.
_EXAMPLE_NBAR = {
    'id': str(_nbar_uuid),
    'product_type': 'nbar_brdf',
    'checksum_path': 'package.sha1',
    'ga_label': 'LS8_OLITIRS_NBAR_P54_GALPGS01-002_112_079_20140126',
    'ga_level': 'P54',
    'size_bytes': 4550,
    'platform': {'code': 'LANDSAT_8'},
    'creation_dt': datetime.datetime(2014, 1, 26, 2, 5, 23, 126373).isoformat(),
    'instrument': {'name': 'OLI_TIRS'},
    'format': {'name': 'GeoTIFF'},
    'extent': {
        'center_dt': datetime.datetime(2014, 1, 26, 2, 5, 23, 126373).isoformat(),
        'coord': {
            'ul': {'lat': -26.37259, 'lon': 116.58914},
            'lr': {'lat': -28.48062, 'lon': 118.96145},
            'ur': {'lat': -26.36025, 'lon': 118.92432},
            'll': {'lat': -28.49412, 'lon': 116.58121}
        }
    },
    'lineage': {
        'machine': {},
        'source_datasets': {
            'ortho': {
                'product_level': 'L1T',
                'product_type': 'ortho',
                'id': str(_ortho_uuid),
                'usgs': {
                    'scene_id': 'LC81120792014026ASA00'
                },
                'extent': {
                    'center_dt': datetime.datetime(2014, 1, 26, 2, 5, 23, 126373).isoformat(),
                    'coord': {
                        'ul': {'lat': -26.37259, 'lon': 116.58914},
                        'lr': {'lat': -28.48062, 'lon': 118.96145},
                        'ur': {'lat': -26.36025, 'lon': 118.92432},
                        'll': {'lat': -28.49412, 'lon': 116.58121}
                    }
                },
                'size_bytes': 1854924494,
                'platform': {
                    'code': 'LANDSAT_8'},
                'creation_dt': datetime.datetime(2015, 4, 7, 0, 58, 8).isoformat(),
                'instrument': {'name': 'OLI_TIRS'},
                'checksum_path': 'package.sha1',
                'ga_label': 'LS8_OLITIRS_OTH_P51_GALPGS01-002_112_079_20140126',
                'grid_spatial': {
                    'projection': {
                        'spatial_reference': 'EPSG:28350',
                        'resampling_option': 'CUBIC_CONVOLUTION',
                        'geo_ref_points': {
                            'ul': {'y': 7082987.5, 'x': 459012.5},
                            'lr': {'y': 6847987.5, 'x': 692012.5},
                            'ur': {'y': 7082987.5, 'x': 692012.5},
                            'll': {'y': 6847987.5, 'x': 459012.5}
                        },
                        'orientation': 'NORTH_UP',
                    }
                },
                'acquisition': {
                    'groundstation': {
                        'code': 'ASA',
                        'eods_domain_code': '002',
                        'label': 'Alice Springs'
                    }
                },
                'format': {'name': 'GEOTIFF'},
                'lineage': {
                    'algorithm': {
                        'name': 'LPGS',
                        'parameters': {},
                        'version': '2.4.0'
                    },
                    'machine': {},
                    'source_datasets': {
                        'satellite_telemetry_data': {
                            'product_type': 'satellite_telemetry_data',
                            'checksum_path': 'package.sha1',
                            'id': str(_telemetry_uuid),
                            'ga_label': 'LS8_OLITIRS_STD-MD_P00_LC81160740742015089ASA00_'
                                        '116_074_20150330T022553Z20150330T022657',

                            'ga_level': 'P00',
                            'size_bytes': 637660782,
                            'platform': {
                                'code': 'LANDSAT_8'},
                            'creation_dt': datetime.datetime(2015, 4, 22, 6, 32, 4).isoformat(),
                            'instrument': {'name': 'OLI_TIRS'},
                            'format': {
                                'name': 'MD'},
                            'lineage': {
                                'source_datasets': {}
                            }
                        }
                    }
                }
            }
        }
    }
}

_EXAMPLE_METADATA_TYPE = MetadataType(
    {
        'name': 'eo',
        'dataset': dict(
            id=['id'],
            label=['ga_label'],
            creation_time=['creation_dt'],
            measurements=['image', 'bands'],
            sources=['lineage', 'source_datasets']
        )
    },
    dataset_search_fields={}
)

_EXAMPLE_DATASET_TYPE = DatasetType(
    _EXAMPLE_METADATA_TYPE,
    {
        'name': 'eo',
        'description': "",
        'metadata_type': 'eo',
        'metadata': {}
    }
)


def _build_dataset(doc):
    sources = {name: _build_dataset(src) for name, src in doc['lineage']['source_datasets'].items()}
    return Dataset(_EXAMPLE_DATASET_TYPE, doc, uris=['file://test.zzz'], sources=sources)


_EXAMPLE_NBAR_DATASET = _build_dataset(_EXAMPLE_NBAR)

DatasetRecord = namedtuple('DatasetRecord', ['id', 'metadata', 'dataset_type_ref', 'uris',
                                             'added', 'added_by', 'archived'])


class MockDb(object):
    def __init__(self):
        self.dataset = {}
        self.dataset_source = set()

    @contextmanager
    def _connect(self):
        yield self

    def begin(self):
        pass

    def commit(self):
        pass

    def rollback(self):
        pass

    def get_dataset(self, id):
        return self.dataset.get(id, None)

    def get_locations(self, dataset):
        return ['file:xxx']

    def datasets_intersection(self, ids):
        return [k for k in ids if k in self.dataset]

    def insert_dataset_location(self, *args, **kwargs):
        return

    def insert_dataset(self, metadata_doc, dataset_id, dataset_type_id):
        # Will we pretend this one was already ingested?
        if dataset_id in self.dataset:
            raise DuplicateRecordError('already ingested')

        self.dataset[dataset_id] = DatasetRecord(dataset_id, deepcopy(metadata_doc), dataset_type_id,
                                                 None, None, None, None)
        return True

    def insert_dataset_source(self, classifier, dataset_id, source_dataset_id):
        self.dataset_source.add((classifier, dataset_id, source_dataset_id))


class MockTypesResource:
    def __init__(self, type_):
        self.type = type_

    def get(self, *args, **kwargs):
        return self.type

    def get_by_name(self, *args, **kwargs):
        return self.type

    @contextmanager
    def _db_connection(self, transaction=False):
        yield MockDb()


class MockIndex:
    def __init__(self, db, product):
        self._db = db
        self.products = MockTypesResource(product)

    def thread_transaction(self):
        return None

    @contextmanager
    def _active_connection(self, transaction=False):
        yield self._db


def test_index_dataset():
    mock_db = MockDb()
    mock_index = MockIndex(mock_db, _EXAMPLE_DATASET_TYPE)
    datasets = DatasetResource(mock_db, mock_index)
    dataset = datasets.add(_EXAMPLE_NBAR_DATASET)

    ids = {d.id for d in mock_db.dataset.values()}
    assert ids == {_nbar_uuid, _ortho_uuid, _telemetry_uuid}

    # Three datasets (ours and the two embedded source datasets)
    assert len(mock_db.dataset) == 3

    # Our three datasets should be linked together
    # Nbar -> Ortho -> Telemetry
    assert len(mock_db.dataset_source) == 2
    assert mock_db.dataset_source == {
        ('ortho', _nbar_uuid, _ortho_uuid),
        ('satellite_telemetry_data', _ortho_uuid, _telemetry_uuid)
    }

    # Nothing ingested, because we reported the first as already ingested.
    dataset = datasets.add(_EXAMPLE_NBAR_DATASET)
    assert len(mock_db.dataset) == 3
    assert len(mock_db.dataset_source) == 2


def test_index_already_ingested_source_dataset():
    mock_db = MockDb()
    mock_index = MockIndex(mock_db, _EXAMPLE_DATASET_TYPE)
    datasets = DatasetResource(mock_db, mock_index)
    dataset = datasets.add(_EXAMPLE_NBAR_DATASET.sources['ortho'])

    assert len(mock_db.dataset) == 2
    assert len(mock_db.dataset_source) == 1

    dataset = datasets.add(_EXAMPLE_NBAR_DATASET)
    assert len(mock_db.dataset) == 3
    assert len(mock_db.dataset_source) == 2


def test_index_two_levels_already_ingested():
    mock_db = MockDb()
    mock_index = MockIndex(mock_db, _EXAMPLE_DATASET_TYPE)
    datasets = DatasetResource(mock_db, mock_index)
    dataset = datasets.add(_EXAMPLE_NBAR_DATASET.sources['ortho'].sources['satellite_telemetry_data'])

    assert len(mock_db.dataset) == 1
    assert len(mock_db.dataset_source) == 0

    dataset = datasets.add(_EXAMPLE_NBAR_DATASET)
    assert len(mock_db.dataset) == 3
    assert len(mock_db.dataset_source) == 2



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
from typing import List

from ._tools import singleton_setup
from .driver_cache import load_drivers


class WriterDriverCache(object):
    def __init__(self, group: str):
        self._drivers = load_drivers(group)

        for driver in list(self._drivers.values()):
            if hasattr(driver, 'aliases'):
                for alias in driver.aliases:
                    self._drivers[alias] = driver

    def __call__(self, name: str):
        """
        :returns: None if driver with a given name is not found

        :param name: Driver name
        :return: Returns WriterDriver
        """
        return self._drivers.get(name, None)

    def drivers(self) -> List[str]:
        """ Returns list of driver names
        """
        return list(self._drivers.keys())


def writer_cache() -> WriterDriverCache:
    """ Singleton for WriterDriverCache
    """
    return singleton_setup(writer_cache, '_instance',
                           WriterDriverCache,
                           'datacube.plugins.io.write')


def writer_drivers() -> List[str]:
    """ Returns list driver names
    """
    return writer_cache().drivers()


def storage_writer_by_name(name):
    """ Lookup writer driver by name

    :returns: Initialised writer driver instance
    :returns: None if driver with this name doesn't exist
    """
    return writer_cache()(name)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Helper methods for working with AWS
"""
import os
import time
import functools
import botocore
import botocore.session
from botocore.credentials import Credentials, ReadOnlyCredentials
from botocore.session import Session
from urllib.request import urlopen
from urllib.parse import urlparse
from sqlalchemy.engine.url import URL

from typing import Optional, Dict, Tuple, Any, Union, IO
from datacube.utils.generic import thread_local_cache

ByteRange = Union[slice, Tuple[int, int]]       # pylint: disable=invalid-name
MaybeS3 = Optional[botocore.client.BaseClient]  # pylint: disable=invalid-name

__all__ = (
    "s3_url_parse",
    "s3_fmt_range",
    "s3_client",
    "s3_open",
    "s3_fetch",
    "s3_dump",
    "ec2_metadata",
    "ec2_current_region",
    "botocore_default_region",
    "auto_find_region",
    "get_creds_with_retry",
    "mk_boto_session",
    "get_aws_settings",
    "configure_s3_access",
)


def _fetch_text(url: str, timeout: float = 0.1) -> Optional[str]:
    try:
        with urlopen(url, timeout=timeout) as resp:
            if 200 <= resp.getcode() < 300:
                return resp.read().decode('utf8')
            else:
                return None
    except IOError:
        return None


def s3_url_parse(url: str) -> Tuple[str, str]:
    """ Return Bucket, Key tuple
    """
    uu = urlparse(url)
    if uu.scheme != "s3":
        raise ValueError("Not a valid s3 url")
    return uu.netloc, uu.path.lstrip('/')


def s3_fmt_range(r: Optional[ByteRange]):
    """ None -> None
        (in, out) -> "bytes={in}-{out-1}"
    """
    if r is None:
        return None

    if isinstance(r, slice):
        if r.step not in [1, None]:
            raise ValueError("Can not process decimated slices")
        if r.stop is None:
            raise ValueError("Can not process open ended slices")

        _in = 0 if r.start is None else r.start
        _out = r.stop
    else:
        _in, _out = r

    if _in < 0 or _out < 0:
        raise ValueError("Slice has to be positive")

    return 'bytes={:d}-{:d}'.format(_in, _out-1)


def ec2_metadata(timeout: float = 0.1) -> Optional[Dict[str, Any]]:
    """ When running inside AWS returns dictionary describing instance identity.
        Returns None when not inside AWS
    """
    import json
    txt = _fetch_text('http://169.254.169.254/latest/dynamic/instance-identity/document', timeout)

    if txt is None:
        return None

    try:
        return json.loads(txt)
    except json.JSONDecodeError:
        return None


def ec2_current_region() -> Optional[str]:
    """ Returns name of the region  this EC2 instance is running in.
    """
    cfg = ec2_metadata()
    if cfg is None:
        return None
    return cfg.get('region', None)


def botocore_default_region(session: Optional[Session] = None) -> Optional[str]:
    """ Returns default region name as configured on the system.
    """
    if session is None:
        session = botocore.session.get_session()
    return session.get_config_variable('region')


def auto_find_region(session: Optional[Session] = None, default: Optional[str] = None) -> str:
    """
    Try to figure out which region name to use

    1. Region as configured for this/default session
    2. Region this EC2 instance is running in
    3. Value supplied in `default`
    4. raise exception
    """
    region_name = botocore_default_region(session)

    if region_name is None:
        region_name = ec2_current_region()

    if region_name is not None:
        return region_name

    if default is None:
        raise ValueError('Region name is not supplied and default can not be found')

    return default


def get_creds_with_retry(session: Session,
                         max_tries: int = 10,
                         sleep: float = 0.1) -> Optional[Credentials]:
    """ Attempt to obtain credentials upto `max_tries` times with back off
    :param session: botocore session, see mk_boto_session
    :param max_tries: number of attempt before failing and returing None
    :param sleep: number of seconds to sleep after first failure (doubles on every consecutive failure)
    """
    for i in range(max_tries):
        if i > 0:
            time.sleep(sleep)
            sleep = min(sleep*2, 10)

        creds = session.get_credentials()
        if creds is not None:
            return creds

    return None


def mk_boto_session(profile: Optional[str] = None,
                    creds: Optional[ReadOnlyCredentials] = None,
                    region_name: Optional[str] = None) -> Session:
    """ Get botocore session with correct `region` configured

    :param profile: profile name to lookup
    :param creds: Override credentials with supplied data
    :param region_name: default region_name to use if not configured for a given profile
    """
    session = botocore.session.Session(profile=profile)

    if creds is not None:
        session.set_credentials(creds.access_key,
                                creds.secret_key,
                                creds.token)

    _region = session.get_config_variable("region")
    if _region is None:
        if region_name is None or region_name == "auto":
            _region = auto_find_region(session, default='us-west-2')
        else:
            _region = region_name
        session.set_config_variable("region", _region)

    return session


def _s3_cache_key(profile: Optional[str] = None,
                  creds: Optional[ReadOnlyCredentials] = None,
                  region_name: Optional[str] = None,
                  aws_unsigned: bool = False,
                  prefix: str = "s3") -> str:
    parts = [prefix,
             "" if creds is None else creds.access_key,
             "T" if aws_unsigned else "F",
             profile or "",
             region_name or ""]
    return ":".join(parts)


def _mk_s3_client(profile: Optional[str] = None,
                  creds: Optional[ReadOnlyCredentials] = None,
                  region_name: Optional[str] = None,
                  session: Optional[Session] = None,
                  use_ssl: bool = True,
                  **cfg) -> botocore.client.BaseClient:
    """ Construct s3 client with configured region_name.

    :param profile    : profile name to lookup (only used if session is not supplied)
    :param creds      : Override credentials with supplied data
    :param region_name: region_name to use, overrides session setting
    :param session    : botocore session to use
    :param use_ssl    : Whether to connect via http or https
    :param cfg        : passed on to ``botocore.client.Config(..)``
                        max_pool_connections
                        connect_timeout
                        read_timeout
                        parameter_validation
                        ...
    """
    if session is None:
        session = mk_boto_session(profile=profile,
                                  creds=creds,
                                  region_name=region_name)

    extras = {}  # type: Dict[str, Any]
    if creds is not None:
        extras.update(aws_access_key_id=creds.access_key,
                      aws_secret_access_key=creds.secret_key,
                      aws_session_token=creds.token)
    if region_name is not None:
        extras['region_name'] = region_name

    return session.create_client('s3',
                                 use_ssl=use_ssl,
                                 **extras,
                                 config=botocore.client.Config(**cfg))


def _aws_unsigned_check_env() -> bool:
    def parse_bool(v: str) -> bool:
        return v.upper() in ('YES', 'Y', 'TRUE', 'T', '1')

    for evar in ('AWS_UNSIGNED', 'AWS_NO_SIGN_REQUEST'):
        v = os.environ.get(evar, None)
        if v is not None:
            return parse_bool(v)

    return False


def s3_client(profile: Optional[str] = None,
              creds: Optional[ReadOnlyCredentials] = None,
              region_name: Optional[str] = None,
              session: Optional[Session] = None,
              aws_unsigned: Optional[bool] = None,
              use_ssl: bool = True,
              cache: Union[bool, str] = False,
              **cfg) -> botocore.client.BaseClient:
    """ Construct s3 client with configured region_name.

    :param profile: profile name to lookup (only used if session is not supplied)
    :param creds: Override credentials with supplied data
    :param region_name: region_name to use, overrides session setting
    :param aws_unsigned: Do not use any credentials when accessing S3 resources
    :param session: botocore session to use
    :param use_ssl: Whether to connect via http or https
    :param cache: ``True`` - store/lookup s3 client in thread local cache.
                  ``"purge"`` - delete from cache and return what was there to begin with

    :param cfg: passed on to ``botocore.client.Config(..)``

    """
    if aws_unsigned is None:
        if creds is None:
            aws_unsigned = _aws_unsigned_check_env()
        else:
            aws_unsigned = False

    if aws_unsigned:
        cfg.update(signature_version=botocore.UNSIGNED)

    if not cache:
        return _mk_s3_client(profile,
                             creds=creds,
                             region_name=region_name,
                             session=session,
                             use_ssl=use_ssl,
                             **cfg)

    _cache = thread_local_cache("__aws_s3_cache", {})

    key = _s3_cache_key(profile=profile,
                        region_name=region_name,
                        creds=creds,
                        aws_unsigned=aws_unsigned)

    if cache == "purge":
        return _cache.pop(key, None)

    s3 = _cache.get(key, None)

    if s3 is None:
        s3 = _mk_s3_client(profile,
                           creds=creds,
                           region_name=region_name,
                           session=session,
                           use_ssl=use_ssl,
                           **cfg)
        _cache[key] = s3

    return s3


def s3_open(url: str,
            s3: MaybeS3 = None,
            range: Optional[ByteRange] = None,  # pylint: disable=redefined-builtin
            **kwargs):
    """ Open whole or part of S3 object

    :param url: s3://bucket/path/to/object
    :param s3: pre-configured s3 client, see make_s3_client()
    :param range: Byte range to read (first_byte, one_past_last_byte), default is whole object
    :param kwargs: are passed on to ``s3.get_object(..)``
    """
    if range is not None:
        try:
            kwargs['Range'] = s3_fmt_range(range)
        except Exception:
            raise ValueError('Bad range passed in: ' + str(range))

    s3 = s3 or s3_client()
    bucket, key = s3_url_parse(url)
    oo = s3.get_object(Bucket=bucket, Key=key, **kwargs)  # type: ignore[attr-defined]
    return oo['Body']


def s3_head_object(url: str,
                   s3: MaybeS3 = None,
                   **kwargs) -> Optional[Dict[str, Any]]:
    """
    Head object, return object metadata.

    :param url: s3://bucket/path/to/object
    :param s3: pre-configured s3 client, see make_s3_client()
    :param kwargs: are passed on to ``s3.head_object(..)``
    """
    from botocore.exceptions import ClientError

    s3 = s3 or s3_client()
    bucket, key = s3_url_parse(url)

    try:
        oo = s3.head_object(Bucket=bucket, Key=key, **kwargs)  # type: ignore[attr-defined]
    except ClientError:
        return None

    meta = oo.pop('ResponseMetadata', {})
    code = meta.get('HTTPStatusCode', 0)
    if 200 <= code < 300:
        return oo

    # it actually raises exceptions when http code is in the "fail" range
    return None  # pragma: no cover


def s3_fetch(url: str,
             s3: MaybeS3 = None,
             range: Optional[ByteRange] = None,  # pylint: disable=redefined-builtin
             **kwargs) -> bytes:
    """ Read entire or part of object into memory and return as bytes

    :param url: s3://bucket/path/to/object
    :param s3: pre-configured s3 client, see make_s3_client()
    :param range: Byte range to read (first_byte, one_past_last_byte), default is whole object
    """
    return s3_open(url, s3=s3, range=range, **kwargs).read()


def s3_dump(data: Union[bytes, str, IO],
            url: str,
            s3: MaybeS3 = None,
            **kwargs):
    """ Write data to s3 object.

    :param data: bytes to write
    :param url: s3://bucket/path/to/object
    :param s3: pre-configured s3 client, see s3_client()
    :param kwargs: Are passed on to ``s3.put_object(..)``

    ContentType
    ACL
    """

    s3 = s3 or s3_client()
    bucket, key = s3_url_parse(url)

    r = s3.put_object(Bucket=bucket,  # type: ignore[attr-defined]
                      Key=key,
                      Body=data,
                      **kwargs)
    code = r['ResponseMetadata']['HTTPStatusCode']
    return 200 <= code < 300


def get_aws_settings(profile: Optional[str] = None,
                     region_name: str = "auto",
                     aws_unsigned: bool = False,
                     requester_pays: bool = False) -> Tuple[Dict[str, Any], Optional[Credentials]]:
    """
    Compute ``aws=`` parameter for ``set_default_rio_config``.

    see also ``datacube.utils.rio.set_default_rio_config``

    Returns a tuple of: ``(aws: Dictionary, creds: session credentials from botocore)``.

    Note that credentials are baked in to ``aws`` setting dictionary,
    however since those might be STS credentials they might require refresh
    hence they are returned from this function separately as well.
    """
    session = mk_boto_session(profile=profile,
                              region_name=region_name)

    region_name = session.get_config_variable("region")

    if aws_unsigned:
        return (dict(region_name=region_name,
                     aws_unsigned=True), None)

    creds = get_creds_with_retry(session)
    if creds is None:
        raise ValueError("Couldn't get credentials")

    cc = creds.get_frozen_credentials()

    return (dict(region_name=region_name,
                 aws_access_key_id=cc.access_key,
                 aws_secret_access_key=cc.secret_key,
                 aws_session_token=cc.token,
                 requester_pays=requester_pays), creds)


def obtain_new_iam_auth_token(url: URL, region_name: str = "auto", profile_name: Optional[str] = None) -> str:
    # Boto3 is not core requirement, but ImportError is probably the right exception to throw anyway.
    from boto3.session import Session as Boto3Session

    session = Boto3Session(profile_name=profile_name)
    client = session.client("rds", region_name=region_name)
    return client.generate_db_auth_token(DBHostname=url.host, Port=url.port, DBUsername=url.username,
                                         Region=region_name)


def configure_s3_access(profile=None,
                        region_name="auto",
                        aws_unsigned=False,
                        requester_pays=False,
                        cloud_defaults=True,
                        client=None,
                        **gdal_opts):
    """ Credentialize for S3 bucket access or configure public access.

    This function obtains credentials for S3 access and passes them on to
    processing threads, either local or on dask cluster.


    .. note::

       if credentials are STS based they will eventually expire, currently
       this case is not handled very well, reads will just start failing
       eventually and will never recover.

    :param profile:        AWS profile name to use
    :param region_name:    Default region_name to use if not configured for a given/default AWS profile
    :param aws_unsigned:   If ``True`` don't bother with credentials when reading from S3
    :param requester_pays: Needed when accessing requester pays buckets

    :param cloud_defaults: Assume files are in the cloud native format, i.e. no side-car files, disables
                           looking for side-car files, makes things faster but won't work for files
                           that do have side-car files with extra metadata.

    :param client:         Dask distributed ``dask.Client`` instance, if supplied apply settings on the
                           dask cluster rather than locally.
    :param gdal_opts:      Any other option to pass to GDAL environment setup

    :returns: credentials object or ``None`` if ``aws_unsigned=True``
    """
    from ..rio import set_default_rio_config

    aws, creds = get_aws_settings(profile=profile,
                                  region_name=region_name,
                                  aws_unsigned=aws_unsigned,
                                  requester_pays=requester_pays)

    if client is None:
        set_default_rio_config(aws=aws, cloud_defaults=cloud_defaults, **gdal_opts)
    else:
        client.register_worker_callbacks(
            functools.partial(set_default_rio_config,
                              aws=aws,
                              cloud_defaults=cloud_defaults,
                              **gdal_opts))

    return creds
